# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.aiding module

"""

import falcon
import json
from dataclasses import asdict, dataclass, field
from typing import Dict, Optional, List, Union
from urllib.parse import urlparse, urljoin
from keri import kering
from keri import core
from keri.app import habbing
from keri.app.keeping import Algos
from keri.core import coring, serdering, eventing
from keri.db import dbing
from keri.help import ogler
from mnemonic import mnemonic
from keri.db import basing
from marshmallow import fields
from marshmallow_dataclass import class_schema

from ..core import longrunning, httping
from ..utils.openapi import namedtupleToEnum, dataclassFromFielddom
from keri.core.serdering import Protocols, Vrsn_1_0, Vrsn_2_0, SerderKERI

logger = ogler.getLogger()


def loadEnds(app, agency, authn):
    groupEnd = AgentResourceEnd(agency=agency, authn=authn)
    app.add_route("/agent/{caid}", groupEnd)

    aidsEnd = IdentifierCollectionEnd()
    app.add_route("/identifiers", aidsEnd)

    aidEnd = IdentifierResourceEnd()
    app.add_route("/identifiers/{name}", aidEnd)
    app.add_route("/identifiers/{name}/events", aidEnd)
    app.add_route("/identifiers/{name}/submit", aidEnd)

    aidOOBIsEnd = IdentifierOOBICollectionEnd()
    app.add_route("/identifiers/{name}/oobis", aidOOBIsEnd)

    endRolesEnd = EndRoleCollectionEnd()
    app.add_route("/identifiers/{name}/endroles", endRolesEnd)
    app.add_route("/identifiers/{name}/endroles/{role}", endRolesEnd)
    app.add_route("/endroles/{aid}", endRolesEnd)
    app.add_route("/endroles/{aid}/{role}", endRolesEnd)

    endRoleEnd = EndRoleResourceEnd()
    app.add_route("/identifiers/{name}/endroles/{role}/{eid}", endRoleEnd)

    locSchemesEnd = LocSchemeCollectionEnd()
    app.add_route("/identifiers/{name}/locschemes", locSchemesEnd)

    rpyEscrowEnd = RpyEscrowCollectionEnd()
    app.add_route("/escrows/rpy", rpyEscrowEnd)

    chaEnd = ChallengeCollectionEnd()
    app.add_route("/challenges", chaEnd)
    chaResEnd = ChallengeResourceEnd()
    app.add_route("/challenges/{name}", chaResEnd)
    chaVerResEnd = ChallengeVerifyResourceEnd()
    app.add_route("/challenges_verify/{source}", chaVerResEnd)

    contactColEnd = ContactCollectionEnd()
    app.add_route("/contacts", contactColEnd)
    contactResEnd = ContactResourceEnd()
    app.add_route("/contacts/{prefix}", contactResEnd)
    contactImgEnd = ContactImageResourceEnd()
    app.add_route("/contacts/{prefix}/img", contactImgEnd)

    groupEnd = GroupMemberCollectionEnd()
    app.add_route("/identifiers/{name}/members", groupEnd)

    return aidEnd


@dataclass
class Seal:
    s: str
    d: str
    i: Optional[str] = field(
        default=None, metadata={"marshmallow_field": fields.String(allow_none=False)}
    )


ixnFieldDomV1 = SerderKERI.Fields[Protocols.keri][Vrsn_1_0][coring.Ilks.ixn]
IXN_V_1, IXNSchema_V_1 = dataclassFromFielddom("IXN_V_1", ixnFieldDomV1)
ixnFieldDomV2 = SerderKERI.Fields[Protocols.keri][Vrsn_2_0][coring.Ilks.ixn]
IXN_V_2, IXNSchema_V_2 = dataclassFromFielddom("IXN_V_2", ixnFieldDomV2)

icpFieldDomV1 = SerderKERI.Fields[Protocols.keri][Vrsn_1_0][coring.Ilks.icp]
ICP_V_1, ICPSchema_V_1 = dataclassFromFielddom("ICP_V_1", icpFieldDomV1)
icpFieldDomV2 = SerderKERI.Fields[Protocols.keri][Vrsn_2_0][coring.Ilks.icp]
ICP_V_2, ICPSchema_V_2 = dataclassFromFielddom("ICP_V_2", icpFieldDomV2)

rotFieldDomV1 = SerderKERI.Fields[Protocols.keri][Vrsn_1_0][coring.Ilks.rot]
ROT_V_1, ROTSchema_V_1 = dataclassFromFielddom("ROT_V_1", rotFieldDomV1)
rotFieldDomV2 = SerderKERI.Fields[Protocols.keri][Vrsn_2_0][coring.Ilks.rot]
ROT_V_2, ROTSchema_V_2 = dataclassFromFielddom("ROT_V_2", rotFieldDomV2)

dipFieldDomV1 = SerderKERI.Fields[Protocols.keri][Vrsn_1_0][coring.Ilks.dip]
DIP_V_1, DIPSchema_V_1 = dataclassFromFielddom("DIP_V_1", dipFieldDomV1)
dipFieldDomV2 = SerderKERI.Fields[Protocols.keri][Vrsn_2_0][coring.Ilks.dip]
DIP_V_2, DIPSchema_V_2 = dataclassFromFielddom("DIP_V_2", dipFieldDomV2)

drtFieldDomV1 = SerderKERI.Fields[Protocols.keri][Vrsn_1_0][coring.Ilks.drt]
DRT_V_1, DRTSchema_V_1 = dataclassFromFielddom("DRT_V_1", drtFieldDomV1)
drtFieldDomV2 = SerderKERI.Fields[Protocols.keri][Vrsn_2_0][coring.Ilks.drt]
DRT_V_2, DRTSchema_V_2 = dataclassFromFielddom("DRT_V_2", drtFieldDomV2)


@dataclass
class KeyStateRecord(basing.KeyStateRecord):
    k: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    n: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    b: list = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    c: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    ee: basing.StateEERecord = field(
        default_factory=basing.StateEERecord,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(basing.StateEERecord), required=True
            )
        },
    )
    kt: Union[str, list[str]]
    nt: Union[str, list[str]]


@dataclass
class Controller:
    state: KeyStateRecord
    ee: Union[
        "ICP_V_1",
        "ICP_V_2",
        "ROT_V_1",
        "ROT_V_2",
        "DIP_V_1",
        "DIP_V_2",
        "DRT_V_1",
        "DRT_V_2",
    ]  # type: ignore


@dataclass
class AgentResourceResult:
    agent: KeyStateRecord
    controller: Controller
    pidx: int
    ridx: Optional[int] = None
    sxlt: Optional[str] = None


class AgentResourceEnd:
    """Resource class for getting agent specific launch information"""

    def __init__(self, agency, authn):
        self.agency = agency
        self.authn = authn

    def on_get(self, _, rep, caid):
        """GET endpoint for Keystores

        Get keystore status

        Args:
            _: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            caid(str): qb64 identifier prefix of Controller

        ---
        summary: Retrieve key state record of an agent by controller AID.
        description: This endpoint retrieves the key state record for a given controller of an agent.
        tags:
        - Agent
        parameters:
        - in: path
          name: caid
          schema:
            type: string
          required: true
          description: The qb64 identifier prefix of Controller.
        responses:
          200:
            description: Successfully retrieved the key state record.
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/AgentResourceResult'
          400:
            description: Bad request. This could be due to an invalid agent or controller configuration.
          404:
            description: The requested controller or agent was not found.
        """
        agent = self.agency.get(caid)
        if agent is None:
            raise falcon.HTTPNotFound(
                description=f"not agent found for controller {caid}"
            )

        if agent.pre not in agent.hby.kevers:
            raise falcon.HTTPBadRequest(
                description=f"invalid agent configuration, {agent.pre} not found"
            )

        if agent.caid not in agent.hby.kevers:
            raise falcon.HTTPBadRequest(
                description=f"invalid controller configuration, {agent.caid} not found"
            )

        pidx = 0
        for name, _ in agent.hby.db.names.getItemIter():
            if name[0] != "agent":
                pidx += 1

        # pidx = agent.hby.db.habs.cntAll()

        state = asdict(agent.hby.kevers[agent.caid].state())
        key = dbing.dgKey(state["i"], state["ee"]["d"])  # digest key
        msg = agent.hby.db.getEvt(key)
        eserder = serdering.SerderKERI(raw=bytes(msg))

        body = dict(
            agent=asdict(agent.hby.kevers[agent.pre].state()),
            controller=dict(state=state, ee=eserder.ked),
            pidx=pidx,
        )

        if (sxlt := agent.mgr.sxlt) is not None:
            body["sxlt"] = sxlt

        rep.content_type = "application/json"
        rep.data = json.dumps(body).encode("utf-8")
        rep.status = falcon.HTTP_200

    def on_put(self, req, rep, caid):
        """

        Parameters:
            req (Request): falcon.Request HTTP request
            rep (Response): falcon.Response HTTP response
            caid(str): qb64 identifier prefix of Controller

        ---
        summary: Update agent configuration by controller AID.
        description: This endpoint updates the agent configuration based on the provided request parameters and body.
        tags:
        - Agent
        parameters:
        - in: path
          name: caid
          schema:
            type: string
          required: true
          description: The qb64 identifier prefix of Controller.
        requestBody:
            required: true
            content:
              application/json:
                schema:
                  type: object
                  required:
                    - rot
                    - sigs
                    - sxlt
                    - kyes
                  properties:
                    rot:
                      type: object
                      description: The rotation event.
                    sigs:
                      type: array
                      items:
                          type: string
                      description: The signatures.
                    sxlt:
                      type: string
                      description: The salty parameters.
                    keys:
                      type: object
                      description: The keys.
        responses:
            204:
              description: Successfully updated the agent configuration.
            400:
              description: Bad request. This could be due to missing or invalid parameters.
            404:
              description: The requested agent was not found.
            500:
              description: Internal server error. This could be due to an issue with updating the agent configuration.
        """
        agent = self.agency.get(caid)
        if agent is None:
            raise falcon.HTTPNotFound(description=f"no agent for {caid}")

        typ = req.params.get("type")
        if typ == "ixn":
            ixn = self.interact(req, rep, agent, caid)
            self.anchorSeals(agent, ixn)
            return

        body = req.get_media()

        if "rot" not in body:
            raise falcon.HTTPBadRequest(
                description="required field 'rot' missing from body"
            )

        if "sigs" not in body:
            raise falcon.HTTPBadRequest(
                description="required field 'sigs' missing from body"
            )

        if "sxlt" not in body:
            raise falcon.HTTPBadRequest(
                description="required field 'sxlt' missing from body"
            )

        if "keys" not in body:
            raise falcon.HTTPBadRequest(
                description="required field 'keys' missing from body"
            )

        rot = serdering.SerderKERI(sad=body["rot"])
        sigs = body["sigs"]

        ctrlHab = agent.hby.habByName(caid, ns="agent")
        ctrlHab.rotate(serder=rot, sigers=[core.Siger(qb64=sig) for sig in sigs])

        if not self.authn.verify(req):
            raise falcon.HTTPForbidden(description="invalid signature on request")

        sxlt = body["sxlt"]
        agent.mgr.sxlt = sxlt

        keys = body["keys"]
        for pre, val in keys.items():
            if "sxlt" in val:
                if (sp := agent.mgr.rb.sprms.get(pre)) is None:
                    raise ValueError(
                        "Attempt to update sxlt for nonexistent or invalid pre={}.".format(
                            pre
                        )
                    )

                sp.sxlt = val["sxlt"]

                if not agent.mgr.rb.sprms.pin(pre, val=sp):
                    raise ValueError(
                        "Unable to update sxlt prms for pre={}.".format(pre)
                    )

            elif "prxs" in val:
                hab = agent.hby.habs[pre]
                verfers = hab.kever.verfers
                digers = hab.kever.digers
                prxs = val["prxs"]

                for idx, prx in enumerate(prxs):
                    cipher = core.Cipher(qb64=prx)
                    agent.mgr.rb.prxs.put(keys=verfers[idx].qb64b, val=cipher)

                if "nxts" in val:
                    nxts = val["nxts"]
                    if len(nxts) != len(digers):
                        raise ValueError(
                            "If encrypted private next keys are provided, must match digers"
                        )

                    for idx, prx in enumerate(nxts):
                        cipher = core.Cipher(qb64=prx)
                        agent.mgr.rb.nxts.put(keys=digers[idx].qb64b, val=cipher)

        agent.mgr.delete_sxlt()

        rep.status = falcon.HTTP_204

    @staticmethod
    def interact(req, rep, agent, caid):
        body = req.get_media()

        if "ixn" not in body:
            raise falcon.HTTPBadRequest(
                description="required field 'ixn' missing from body"
            )

        if "sigs" not in body:
            raise falcon.HTTPBadRequest(
                description="required field 'sigs' missing from body"
            )

        ked = body["ixn"]
        sigs = body["sigs"]
        ixn = serdering.SerderKERI(sad=ked)
        sigers = [core.Siger(qb64=sig) for sig in sigs]

        ctrlHab = agent.hby.habByName(caid, ns="agent")

        ctrlHab.interact(serder=ixn, sigers=sigers)
        agent.agentHab.kvy.processEvent(serder=ixn, sigers=sigers)

        rep.status = falcon.HTTP_204

        return ixn

    @staticmethod
    def anchorSeals(agent, ixn):
        a = ixn.ked["a"]
        if len(a) == 0:
            return

        delegator = coring.Saider(qb64=ixn.ked["d"])
        delegatorsn = coring.Seqner(snh=ixn.ked["s"])

        seal = a[0]
        prefixer = coring.Prefixer(qb64=seal["i"])
        saider = coring.Saider(qb64=seal["d"])

        couple = delegatorsn.qb64b + delegator.qb64b
        dgkey = dbing.dgKey(prefixer.qb64b, saider.qb64)
        agent.hby.db.setAes(dgkey, couple)  # authorizer event seal (delegator/issuer)


@dataclass
class RandyKeyState:
    prxs: List[str]
    nxts: List[str]


@dataclass
class ExternState:
    extern_type: str
    pidx: int
    # Override the schema to force additionalProperties=True


Tier = namedtupleToEnum(coring.Tiers, "Tier")


@dataclass
class SaltyState:
    tier: Tier  # type: ignore
    sxlt: str = ""
    pidx: int = 0
    kidx: int = 0
    stem: str = ""
    dcode: str = ""
    icodes: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    ncodes: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    transferable: bool = False


@dataclass
class GroupKeyState:
    """Data class for group key state"""

    mhab: "HabState"
    keys: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )
    ndigs: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )


@dataclass
class HabStateBase:
    """Base data class for identifier state (minimal)"""

    name: str
    prefix: str
    icp_dt: str
    # One of salty, randy, group, or extern must be present
    # Patch to ensure only one of these is set in specing

    def __post_init__(self):
        present = [self.salty, self.randy, self.group, self.extern]
        if sum(x is not None for x in present) != 1:
            raise ValueError(
                "Exactly one of salty, randy, group, or extern must be present."
            )


@dataclass
class HabState(HabStateBase):
    """Data class for identifier resource result (full state)"""

    state: KeyStateRecord = field(
        default_factory=KeyStateRecord,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(KeyStateRecord), required=False
            )
        },
    )
    transferable: bool = field(
        default=False,
        metadata={"marshmallow_field": fields.Boolean(required=False)},
    )
    windexes: list[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=False)},
    )


class IdentifierCollectionEnd:
    """Resource class for creating and managing identifiers"""

    @staticmethod
    def on_options(req, rep):
        rep.add_header("Accept-Ranges", "aids")
        rep.status = falcon.HTTP_200

    @staticmethod
    def on_get(req, rep):
        """Identifier List GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        ---
        summary: Retrieve a list of identifiers associated with the agent.
        description: This endpoint retrieves a list of identifiers associated with the agent.
                     It supports pagination through the 'Range' header.
        tags:
          - Identifier
        parameters:
        - in: header
          name: Range
          schema:
            type: string
          required: false
          description: The 'Range' header is used for pagination. The default range is 0-9.
        responses:
            200:
                description: Successfully retrieved identifiers.
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        $ref: '#/components/schemas/HabStateBase'
            206:
                description: Successfully retrieved identifiers within the specified range.
        """
        agent = req.context.agent
        res = []

        rng = req.get_header("Range")
        if rng is None:
            rep.status = falcon.HTTP_200
            start = 0
            end = 9
        else:
            rep.status = falcon.HTTP_206
            start, end = httping.parseRangeHeader(rng, "aids")

        count = agent.hby.db.habs.cntAll()
        it = agent.hby.db.names.getItemIter()
        for _ in range(start):
            try:
                next(it)
            except StopIteration:
                break

        for (ns, name), pre in it:
            if ns == "agent":
                continue

            hab = agent.hby.habs[pre]
            data = info(hab, agent.mgr)
            res.append(data)

            if (not end == -1) and len(res) == (end - start) + 1:
                break

        end = start + (len(res) - 1) if len(res) > 0 else 0
        rep.set_header("Accept-Ranges", "aids")
        rep.set_header("Content-Range", f"aids {start}-{end}/{count - 1}")
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")

    @staticmethod
    def on_post(req, rep):
        """Inception event POST endpoint

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object

        ---
        summary: Create an identifier.
        description: This endpoint creates an identifier with the provided inception event, name, and signatures.
        tags:
        - Identifier
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    icp:
                      type: object
                      description: The inception event for the identifier.
                    name:
                      type: string
                      description: The name of the identifier.
                    sigs:
                      type: array
                      items:
                          type: string
                      description: The signatures for the inception event.
                    group:
                      type: object
                      description: Multisig group information.
                    salty:
                      type: object
                      description: Salty parameters.
                    randy:
                      type: object
                      description: Randomly generated materials.
                    extern:
                      type: object
                      description: External parameters.
        responses:
            202:
                description: Identifier creation is in progress. The response is a long running operation.
                content:
                  application/json:
                    schema:
                      oneOf:
                        - $ref: '#/components/schemas/GroupOperation'
                        - $ref: '#/components/schemas/WitnessOperation'
                        - $ref: '#/components/schemas/DelegationOperation'
                        - $ref: '#/components/schemas/DoneOperation'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
        """
        agent = req.context.agent
        try:
            body = req.get_media()
            icp = httping.getRequiredParam(body, "icp")
            name = httping.getRequiredParam(body, "name")
            sigs = httping.getRequiredParam(body, "sigs")

            serder = serdering.SerderKERI(sad=icp)

            sigers = [core.Siger(qb64=sig) for sig in sigs]

            if agent.hby.habByName(name) is not None:
                raise falcon.HTTPBadRequest(
                    title=f"AID with name {name} already incepted"
                )

            if "b" in icp:
                for wit in icp["b"]:
                    urls = agent.agentHab.fetchUrls(eid=wit, scheme=kering.Schemes.http)
                    if not urls and wit not in agent.hby.kevers:
                        raise falcon.HTTPBadRequest(
                            description=f"unknown witness {wit}"
                        )

            if "di" in icp and icp["di"] not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(
                    description=f"unknown delegator {icp['di']}"
                )

            # client is requesting agent to join multisig group
            if "group" in body:
                group = body["group"]

                if "mhab" not in group:
                    raise falcon.HTTPBadRequest(
                        description='required field "mhab" missing from body.group'
                    )
                mpre = group["mhab"]["prefix"]

                if mpre not in agent.hby.habs:
                    raise falcon.HTTPBadRequest(
                        description=f"signing member {mpre} not a local AID"
                    )
                mhab = agent.hby.habs[mpre]

                if "keys" not in group:
                    raise falcon.HTTPBadRequest(
                        description='required field "keys" missing from body.group'
                    )
                keys = group["keys"]
                verfers = [coring.Verfer(qb64=key) for key in keys]

                if mhab.kever.fetchLatestContribTo(verfers=verfers) is None:
                    raise falcon.HTTPBadRequest(
                        description=f"Member hab={mhab.pre} not a participant in "
                        f"event for this group hab."
                    )

                if "ndigs" not in group:
                    raise falcon.HTTPBadRequest(
                        description='required field "ndigs" missing from body.group'
                    )
                ndigs = group["ndigs"]
                digers = [coring.Diger(qb64=ndig) for ndig in ndigs]

                states = httping.getRequiredParam(body, "smids")
                rstates = httping.getRequiredParam(body, "rmids")

                hab = agent.hby.makeSignifyGroupHab(
                    name,
                    mhab=mhab,
                    smids=states,
                    rmids=rstates,
                    serder=serder,
                    sigers=sigers,
                )
                try:
                    agent.inceptGroup(
                        pre=serder.pre, mpre=mhab.pre, verfers=verfers, digers=digers
                    )
                except ValueError as e:
                    agent.hby.deleteHab(name=name)
                    raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                # Generate response, a long running operaton indicator for the type
                agent.groups.append(
                    dict(
                        pre=hab.pre,
                        serder=serder,
                        sigers=sigers,
                        smids=states,
                        rmids=rstates,
                    )
                )
                op = agent.monitor.submit(
                    serder.pre,
                    longrunning.OpTypes.group,
                    metadata=dict(pre=hab.pre, sn=0),
                )

                rep.content_type = "application/json"
                rep.status = falcon.HTTP_202
                rep.data = op.to_json().encode("utf-8")

            else:
                # client is requesting that the Agent track the Salty parameters
                if Algos.salty in body:
                    salt = body[Algos.salty]
                    hab = agent.hby.makeSignifyHab(name, serder=serder, sigers=sigers)
                    try:
                        agent.inceptSalty(pre=serder.pre, **salt)
                    except ValueError as e:
                        agent.hby.deleteHab(name=name)
                        raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                # client is storing encrypted randomly generated key material on agent
                elif Algos.randy in body:
                    rand = body[Algos.randy]
                    hab = agent.hby.makeSignifyHab(name, serder=serder, sigers=sigers)
                    try:
                        agent.inceptRandy(
                            pre=serder.pre,
                            verfers=serder.verfers,
                            digers=serder.ndigers,
                            **rand,
                        )
                    except ValueError as e:
                        agent.hby.deleteHab(name=name)
                        raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                elif Algos.extern in body:
                    extern = body[Algos.extern]
                    hab = agent.hby.makeSignifyHab(name, serder=serder, sigers=sigers)
                    try:
                        agent.inceptExtern(
                            pre=serder.pre,
                            verfers=serder.verfers,
                            digers=serder.ndigers,
                            **extern,
                        )
                    except ValueError as e:
                        agent.hby.deleteHab(name=name)
                        raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                else:
                    raise falcon.HTTPBadRequest(
                        description="invalid request: one of group, rand or salt field required"
                    )

                # create Hab and incept the key store (if any)
                # Generate response, either the serder or a long running operaton indicator for the type
                rep.content_type = "application/json"
                if hab.kever.delpre:
                    agent.anchors.append(dict(pre=hab.pre, sn=0))
                    op = agent.monitor.submit(
                        hab.kever.prefixer.qb64,
                        longrunning.OpTypes.delegation,
                        metadata=dict(pre=hab.pre, sn=0),
                    )
                    rep.status = falcon.HTTP_202
                    rep.data = op.to_json().encode("utf-8")

                elif hab.kever.wits:
                    agent.witners.append(dict(serder=serder))
                    op = agent.monitor.submit(
                        hab.kever.prefixer.qb64,
                        longrunning.OpTypes.witness,
                        metadata=dict(pre=hab.pre, sn=0),
                    )
                    rep.status = falcon.HTTP_202
                    rep.data = op.to_json().encode("utf-8")

                else:
                    rep.status = falcon.HTTP_202
                    op = agent.monitor.submit(
                        hab.kever.prefixer.qb64,
                        longrunning.OpTypes.done,
                        metadata=dict(pre=hab.pre, response=serder.ked),
                    )
                    rep.data = op.to_json().encode("utf-8")

        except (kering.AuthError, ValueError) as e:
            raise falcon.HTTPBadRequest(description=e.args[0])


class IdentifierResourceEnd:
    """Resource class for updating and deleting identifiers"""

    @staticmethod
    def on_get(req, rep, name):
        """Identifier GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human-readable name for Hab to GET

        ---
        summary: Retrieve an identifier.
        description: This endpoint retrieves an identifier by its prefix or human-readable name.
        tags:
        - Identifier
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: true
          description: The human-readable name of the identifier or its prefix.
        responses:
            200:
                description: Successfully retrieved the identifier details.
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/HabState'
            400:
                description: Bad request. This could be due to a missing or invalid name parameter.
            404:
                description: The requested identifier was not found.
        """
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")

        agent = req.context.agent
        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.HTTPNotFound(
                description=f"{name} is not a valid identifier name or prefix"
            )

        data = info(hab, agent.mgr, full=True)
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(data).encode("utf-8")

    def on_put(self, req, rep, name):
        """Identifier rename endpoint

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object
            name (str): human readable name for Hab to rename or its prefix

        ---
        summary: Rename an identifier.
        description: This endpoint renames an identifier with the provided new name.
        tags:
        - Identifier
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: true
          description: The current human-readable name of the identifier or its prefix.
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    name:
                      type: string
                      description: The new human-readable name for the identifier.
                  required:
                  - name
        responses:
            200:
              description: Successfully renamed the identifier and returns the updated information.
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/HabState'
            400:
              description: Bad request. This could be due to a missing or invalid name parameter.
            404:
              description: The requested identifier was not found.
        """
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")

        body = req.get_media()
        newName = body.get("name")

        if not newName:
            raise falcon.HTTPBadRequest(description="new name is required")

        agent = req.context.agent

        if agent.hby.habByName(newName) is not None:
            raise falcon.HTTPBadRequest(
                title=f"AID with name {newName} already incepted"
            )

        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )

        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID with name or prefix {name} found")

        habord = hab.db.habs.get(keys=(hab.pre,))
        oldName = habord.name
        habord.name = newName
        hab.db.habs.pin(keys=(hab.pre,), val=habord)
        hab.db.names.pin(keys=("", newName), val=hab.pre)
        hab.db.names.rem(keys=("", oldName))
        hab.name = newName
        hab = agent.hby.habByName(newName)
        data = info(hab, agent.mgr, full=True)
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(data).encode("utf-8")

    def on_post(self, req, rep, name):
        """Identifier events endpoint

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object
            name (str): human-readable name or prefix for Hab to rotate or interact

        ---
        summary: Process identifier events.
        description: This endpoint handles the 'rot' or 'ixn' events of an identifier, or the request to resubmit the KEL, based on the provided request.
        tags:
        - Identifier
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: true
          description: The human-readable name of the identifier or its prefix.
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    rot:
                      type: object
                      description: The rotation event details.
                    ixn:
                      type: object
                      description: The interaction event details.
                    submit:
                      type: object
                      description: The request to resubmit event details to witnesses.
                  oneOf:
                  - required:
                    - rot
                  - required:
                    - ixn
                  - required:
                    - submit
        responses:
            200:
              description: Successfully processed the identifier's event.
              content:
                application/json:
                  schema:
                      oneOf:
                        - $ref: '#/components/schemas/GroupOperation'
                        - $ref: '#/components/schemas/WitnessOperation'
                        - $ref: '#/components/schemas/DelegationOperation'
                        - $ref: '#/components/schemas/DoneOperation'
                        - $ref: '#/components/schemas/SubmitOperation'
            400:
              description: Bad request. This could be due to missing or invalid parameters.
        """
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")
        agent = req.context.agent
        try:
            body = req.get_media()
            if body.get("rot") is not None:
                op = self.rotate(agent, name, body)
            elif body.get("ixn") is not None:
                op = self.interact(agent, name, body)
            elif body.get("submit") is not None:
                op = self.submit_id(agent, name, body)
            else:
                raise falcon.HTTPBadRequest(
                    title="invalid request",
                    description="required field 'rot' or 'ixn' missing from request",
                )

            rep.status = falcon.HTTP_200
            rep.content_type = "application/json"
            rep.data = op.to_json().encode("utf-8")

        except (kering.AuthError, ValueError) as e:
            raise falcon.HTTPBadRequest(description=e.args[0])

    @staticmethod
    def rotate(agent, name, body):
        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID with name or prefix {name} found")

        rot = body.get("rot")
        if rot is None:
            raise falcon.HTTPBadRequest(
                title="invalid rotation",
                description="required field 'rot' missing from request",
            )
        serder = serdering.SerderKERI(sad=rot)
        logger.info(
            "[%s | %s]: Rotation event sn=%s SAID=%s",
            hab.name,
            hab.pre,
            serder.sn,
            serder.said,
        )

        if "ba" in rot:
            for wit in rot["ba"]:
                urls = agent.agentHab.fetchUrls(eid=wit, scheme=kering.Schemes.http)
                if not urls and wit not in agent.hby.kevers:
                    raise falcon.HTTPBadRequest(description=f"unknown witness {wit}")

        sigs = body.get("sigs")
        if sigs is None or len(sigs) == 0:
            raise falcon.HTTPBadRequest(
                title="invalid rotation",
                description="required field 'sigs' missing from request",
            )
        sigers = [core.Siger(qb64=sig) for sig in sigs]

        if Algos.salty in body:
            hab.rotate(serder=serder, sigers=sigers)

            salt = body[Algos.salty]
            keeper = agent.mgr.get(Algos.salty)

            try:
                keeper.rotate(pre=serder.pre, **salt)
            except ValueError as e:
                agent.hby.deleteHab(name=name)
                raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

        elif Algos.randy in body:
            hab.rotate(serder=serder, sigers=sigers)

            rand = body[Algos.randy]
            keeper = agent.mgr.get(Algos.randy)

            keeper.rotate(
                pre=serder.pre, verfers=serder.verfers, digers=serder.ndigers, **rand
            )

        elif Algos.group in body:
            smids = httping.getRequiredParam(body, "smids")
            rmids = httping.getRequiredParam(body, "rmids")

            hab.rotate(serder=serder, sigers=sigers, smids=smids, rmids=rmids)

            keeper = agent.mgr.get(Algos.group)

            keeper.rotate(pre=serder.pre, verfers=serder.verfers, digers=serder.ndigers)

            agent.groups.append(
                dict(
                    pre=hab.pre, serder=serder, sigers=sigers, smids=smids, rmids=rmids
                )
            )
            op = agent.monitor.submit(
                serder.said,
                longrunning.OpTypes.group,
                metadata=dict(pre=hab.pre, sn=serder.sn),
            )

            return op

        if hab.kever.delpre:
            agent.anchors.append(dict(alias=name, pre=hab.pre, sn=serder.sn))
            op = agent.monitor.submit(
                serder.said,
                longrunning.OpTypes.delegation,
                metadata=dict(pre=hab.pre, sn=serder.sn),
            )
            return op

        if hab.kever.wits:
            agent.witners.append(dict(serder=serder))
            op = agent.monitor.submit(
                serder.said,
                longrunning.OpTypes.witness,
                metadata=dict(pre=hab.pre, sn=serder.sn),
            )
            return op

        op = agent.monitor.submit(
            serder.said,
            longrunning.OpTypes.done,
            metadata=dict(response=serder.ked),
        )
        return op

    @staticmethod
    def interact(agent, name, body):
        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID {name} found")

        ixn = body.get("ixn")
        if ixn is None:
            raise falcon.HTTPBadRequest(
                title="invalid interaction",
                description="required field 'ixn' missing from request",
            )
        serder = serdering.SerderKERI(sad=ixn)
        logger.info(
            "[%s | %s] Interaction event sn=%s SAID=%s",
            hab.name,
            hab.pre,
            serder.sn,
            serder.said,
        )

        sigs = body.get("sigs")
        if sigs is None or len(sigs) == 0:
            raise falcon.HTTPBadRequest(
                title="invalid interaction",
                description="required field 'sigs' missing from request",
            )

        sigers = [core.Siger(qb64=sig) for sig in sigs]

        hab.interact(serder=serder, sigers=sigers)

        if "group" in body:
            agent.groups.append(dict(pre=hab.pre, serder=serder, sigers=sigers))
            op = agent.monitor.submit(
                serder.said,
                longrunning.OpTypes.group,
                metadata=dict(pre=hab.pre, sn=serder.sn),
            )

            return op

        if hab.kever.wits:
            agent.witners.append(dict(serder=serder))
            op = agent.monitor.submit(
                serder.said,
                longrunning.OpTypes.witness,
                metadata=dict(pre=hab.pre, sn=serder.sn),
            )
            return op

        op = agent.monitor.submit(
            serder.said,
            longrunning.OpTypes.done,
            metadata=dict(response=serder.ked),
        )
        return op

    @staticmethod
    def submit_id(agent, name, body):
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID {name} found")

        code = body.get("code")
        logger.info("[%s | %]: Resubmit event code=%s", name, hab.pre, code)

        if hab.kever.wits:
            agent.submits.append(dict(alias=name, code=code))
            op = agent.monitor.submit(
                hab.kever.prefixer.qb64,
                longrunning.OpTypes.submit,
                metadata=dict(alias=name, sn=hab.kever.sn),
            )
            return op

        raise falcon.HTTPBadRequest(
            title=f"invalid identifier submitted, {name} has no witnesses"
        )


def info(hab, rm, full=False):
    data = dict(
        name=hab.name,
        prefix=hab.pre,
    )

    if not isinstance(hab, habbing.SignifyHab) and not isinstance(
        hab, habbing.SignifyGroupHab
    ):
        raise kering.ConfigurationError(
            f"agent only allows SignifyHab instances, {type(hab)}"
        )

    keeper = rm.get(pre=hab.pre)
    data.update(keeper.params(pre=hab.pre))
    if isinstance(hab, habbing.SignifyGroupHab):
        data["group"]["mhab"] = info(hab.mhab, rm, full)

    data["icp_dt"] = bytes(hab.db.getDts(eventing.dgKey(hab.pre, hab.pre))).decode(
        "utf-8"
    )

    if hab.accepted and full:
        kever = hab.kevers[hab.pre]
        data["transferable"] = kever.transferable
        data["state"] = asdict(kever.state())
        dgkey = dbing.dgKey(kever.prefixer.qb64b, kever.serder.saidb)
        wigs = hab.db.getWigs(dgkey)
        data["windexes"] = [core.Siger(qb64b=bytes(wig)).index for wig in wigs]

    return data


Role = namedtupleToEnum(kering.Roles, "Role")


@dataclass
class OOBI:
    """Data class for OOBI URLs"""

    role: Role  # type: ignore
    oobis: List[str] = field(
        default_factory=list,
        metadata={"marshmallow_field": fields.List(fields.String(), required=True)},
    )


class IdentifierOOBICollectionEnd:
    """
    This class represents the OOBI subresource collection endpoint for identifiers

    """

    @staticmethod
    def on_get(req, rep, name):
        """Identifier GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human-readable name or prefix for Hab to GET
        ---
        summary: Fetch OOBI URLs of an identifier.
        description: This endpoint fetches the OOBI URLs for a specific role associated with an identifier.
        tags:
        - Identifier
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: true
          description: The human-readable name of the identifier or its prefix.
        - in: query
          name: role
          schema:
            type: string
          required: true
          description: The role for which to fetch the OOBI URLs. Can be a witness, controller, agent, or mailbox.
        responses:
            200:
              description: Successfully fetched the OOBI URLs. The response body contains the OOBI URLs.
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/OOBI'
            400:
              description: Bad request. This could be due to missing or invalid parameters.
            404:
              description: The requested identifier was not found.
        """
        agent = req.context.agent
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")

        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if not hab:
            raise falcon.HTTPNotFound(description="invalid alias or prefix {name}")

        if "role" not in req.params:
            raise falcon.HTTPBadRequest(description="role parameter required")

        role = req.params["role"]

        res = dict(role=role)
        if role in (kering.Roles.witness,):  # Fetch URL OOBIs for all witnesses
            oobis = []
            for wit in hab.kever.wits:
                urls = hab.fetchUrls(
                    eid=wit, scheme=kering.Schemes.http
                ) or hab.fetchUrls(eid=wit, scheme=kering.Schemes.https)
                if not urls:
                    raise falcon.HTTPNotFound(
                        description=f"unable to query witness {wit}, no http endpoint"
                    )

                url = (
                    urls[kering.Schemes.http]
                    if kering.Schemes.http in urls
                    else urls[kering.Schemes.https]
                )
                up = urlparse(url)
                oobis.append(urljoin(up.geturl(), f"/oobi/{hab.pre}/witness/{wit}"))
            res["oobis"] = oobis
        elif role in (kering.Roles.controller,):  # Fetch any controller URL OOBIs
            oobis = []
            urls = hab.fetchUrls(
                eid=hab.pre, scheme=kering.Schemes.http
            ) or hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.https)
            if not urls:
                raise falcon.HTTPNotFound(
                    description=f"unable to query controller {hab.pre}, no http endpoint"
                )

            url = (
                urls[kering.Schemes.http]
                if kering.Schemes.http in urls
                else urls[kering.Schemes.https]
            )
            up = urlparse(url)
            oobis.append(urljoin(up.geturl(), f"/oobi/{hab.pre}/controller"))
            res["oobis"] = oobis
        elif role in (kering.Roles.agent,):  # Fetch URL OOBIs for all witnesses
            roleUrls = hab.fetchRoleUrls(
                cid=hab.pre, role=kering.Roles.agent, scheme=kering.Schemes.http
            ) or hab.fetchRoleUrls(
                cid=hab.pre, role=kering.Roles.agent, scheme=kering.Schemes.https
            )
            if kering.Roles.agent not in roleUrls:
                res["oobis"] = []
            else:
                aoobis = roleUrls[kering.Roles.agent]

                oobis = list()
                for agent in set(aoobis.keys()):
                    murls = aoobis.naball(agent)
                    for murl in murls:
                        urls = []
                        if kering.Schemes.http in murl:
                            urls.extend(murl.naball(kering.Schemes.http))
                        if kering.Schemes.https in murl:
                            urls.extend(murl.naball(kering.Schemes.https))
                        for url in urls:
                            up = urlparse(url)
                            oobis.append(
                                urljoin(up.geturl(), f"/oobi/{hab.pre}/agent/{agent}")
                            )

                res["oobis"] = oobis
        elif role in (kering.Roles.mailbox,):  # Fetch URL OOBIs for all witnesses
            roleUrls = hab.fetchRoleUrls(
                cid=hab.pre, role=kering.Roles.mailbox, scheme=kering.Schemes.http
            ) or hab.fetchRoleUrls(
                cid=hab.pre, role=kering.Roles.mailbox, scheme=kering.Schemes.https
            )
            if kering.Roles.mailbox not in roleUrls:
                res["oobis"] = []
            else:
                aoobis = roleUrls[kering.Roles.mailbox]

                oobis = list()
                for mailbox in set(aoobis.keys()):
                    murls = aoobis.naball(mailbox)
                    for murl in murls:
                        urls = []
                        if kering.Schemes.http in murl:
                            urls.extend(murl.naball(kering.Schemes.http))
                        if kering.Schemes.https in murl:
                            urls.extend(murl.naball(kering.Schemes.https))
                        for url in urls:
                            up = urlparse(url)
                            oobis.append(
                                urljoin(
                                    up.geturl(), f"/oobi/{hab.pre}/mailbox/{mailbox}"
                                )
                            )

                res["oobis"] = oobis
        else:
            raise falcon.HTTPBadRequest(
                description=f"unsupport role type {role} for oobi request"
            )

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")


@dataclass
class EndRole:
    """Data class for End Role"""

    cid: str
    role: str
    eid: str


class EndRoleCollectionEnd:
    @staticmethod
    def on_get(req, rep, name=None, aid=None, role=None):
        """GET endpoint for end role collection

        Parameters:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object
            name (str): human readable alias or prefix for AID
            aid (str): aid to use instead of name
            aid (str): aid to use instead of name
            role (str): optional role to search for

        ---
        summary: Retrieve end roles.
        description: This endpoint retrieves the end roles associated with an identifier prefix or human-readable name.
                     It can also filter the end roles based on a specific role.
        tags:
        - End Role
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: false
          description: The human-readable name of the identifier or its prefix.
        - in: path
          name: aid
          schema:
            type: string
          required: false
          description: The identifier (AID).
        - in: path
          name: role
          schema:
            type: string
          required: false
          description: The specific role to filter the end roles.
        responses:
            200:
                description: Successfully retrieved the end roles. The response body contains the end roles.
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        $ref: '#/components/schemas/EndRole'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
            404:
                description: The requested identifier was not found.
        """
        agent = req.context.agent

        if name is not None:
            hab = (
                agent.hby.habs[name]
                if name in agent.hby.habs
                else agent.hby.habByName(name)
            )
            if hab is None:
                raise falcon.errors.HTTPNotFound(
                    description=f"invalid alias or prefix {name}"
                )
            pre = hab.pre
        elif aid is not None:
            pre = aid
        else:
            raise falcon.HTTPBadRequest(
                description="name or prefix is required in the path"
            )

        if role is not None:
            keys = (
                pre,
                role,
            )
        else:
            keys = (pre,)

        ends = []
        for (_, erole, eid), end in agent.hby.db.ends.getItemIter(keys=keys):
            ends.append(dict(cid=pre, role=erole, eid=eid))

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(ends).encode("utf-8")

    @staticmethod
    def on_post(req, rep, name, aid=None, role=None):
        """POST endpoint for end role collection

        Args:
            req (Request): Falcon HTTP request object
            rep (Response): Falcon HTTP response object
            name (str): human readable alias or prefix for identifier
            aid (str): Not supported for POST.  If provided, a 404 is returned
            role (str): Not supported for POST.  If provided, a 404 is returned

        ---
        summary: Create an end role.
        description: This endpoint creates an end role associated with a given identifier prefix or human-readable name.
        tags:
        - End Role
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: true
          description: The human-readable name of the identifier or its prefix.
        - in: path
          name: aid
          schema:
            type: string
          required: false
          description: Not supported for POST. If provided, a 404 is returned.
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    rpy:
                      type: object
                      description: The reply object.
                    sigs:
                      type: array
                      items:
                        type: string
                      description: The signatures.
        responses:
            202:
                description: Accepted. The end role creation is in progress.
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/EndRoleOperation'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
            404:
                description: Not found. The requested identifier was not found.
        """
        if role is not None or aid is not None:
            raise falcon.HTTPNotFound(description="route not found")

        agent = req.context.agent
        body = req.get_media()

        rpy = httping.getRequiredParam(body, "rpy")
        rsigs = httping.getRequiredParam(body, "sigs")

        rserder = serdering.SerderKERI(sad=rpy)
        data = rserder.ked["a"]
        pre = data["cid"]
        role = data["role"]
        eid = data["eid"]

        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.errors.HTTPNotFound(
                description=f"invalid alias or prefix {name}"
            )

        if pre != hab.pre:
            raise falcon.errors.HTTPBadRequest(
                description=f"error trying to create end role for unknown local AID {pre}"
            )

        rsigers = [core.Siger(qb64=rsig) for rsig in rsigs]
        tsg = (
            hab.kever.prefixer,
            coring.Seqner(sn=hab.kever.sn),
            coring.Saider(qb64=hab.kever.serder.said),
            rsigers,
        )
        try:
            agent.hby.rvy.processReply(rserder, tsgs=[tsg])
        except kering.UnverifiedReplyError:
            pass
        except kering.ValidationError:
            raise falcon.HTTPBadRequest(
                description="unable to verify end role reply message"
            )

        oid = ".".join([pre, role, eid])
        op = agent.monitor.submit(
            oid, longrunning.OpTypes.endrole, metadata=dict(cid=pre, role=role, eid=eid)
        )

        rep.content_type = "application/json"
        rep.status = falcon.HTTP_202
        rep.data = op.to_json().encode("utf-8")


class EndRoleResourceEnd:
    def on_delete(self, req, rep):
        pass


class LocSchemeCollectionEnd:
    @staticmethod
    def on_post(req, rep, name):
        """POST endpoint for loc scheme collection

        Args:
            req (Request): Falcon HTTP request object
            rep (Response): Falcon HTTP response object
            name (str): human readable alias or prefix for identifier

        ---
        summary: Authorises a new location scheme.
        description: This endpoint authorises a new location scheme (endpoint) for a particular endpoint identifier.
        tags:
        - Loc Scheme
        parameters:
        - in: path
          name: name
          schema:
            type: string
          required: true
          description: The human-readable name of the identifier or its prefix.
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    rpy:
                      type: object
                      description: The reply object.
                    sigs:
                      type: array
                      items:
                        type: string
                      description: The signatures.
        responses:
            202:
                description: Accepted. The loc scheme authorisation is in progress.
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/LocSchemeOperation'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
            404:
                description: Not found. The requested identifier was not found.
        """
        agent = req.context.agent
        body = req.get_media()

        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.errors.HTTPNotFound(
                description=f"invalid alias or prefix {name}"
            )

        rpy = httping.getRequiredParam(body, "rpy")
        rsigs = httping.getRequiredParam(body, "sigs")

        rserder = serdering.SerderKERI(sad=rpy)
        data = rserder.ked["a"]
        eid = data["eid"]
        scheme = data["scheme"]
        url = data["url"]

        rsigers = [core.Siger(qb64=rsig) for rsig in rsigs]
        tsg = (
            hab.kever.prefixer,
            coring.Seqner(sn=hab.kever.sn),
            coring.Saider(qb64=hab.kever.serder.said),
            rsigers,
        )
        try:
            agent.hby.rvy.processReply(rserder, tsgs=[tsg])
        except kering.UnverifiedReplyError:
            pass
        except kering.ValidationError:
            raise falcon.HTTPBadRequest(
                description="unable to verify end role reply message"
            )

        oid = ".".join([eid, scheme])
        op = agent.monitor.submit(
            oid,
            longrunning.OpTypes.locscheme,
            metadata=dict(eid=eid, scheme=scheme, url=url),
        )

        rep.content_type = "application/json"
        rep.status = falcon.HTTP_202
        rep.data = op.to_json().encode("utf-8")


rpyFieldDomV1 = serdering.SerderKERI.Fields[kering.Protocols.keri][kering.Vrsn_1_0][
    kering.Ilks.rpy
]
RPY_V_1, RpySchema_V_1 = dataclassFromFielddom("RPY_V_1", rpyFieldDomV1)
rpyFieldDomV2 = serdering.SerderKERI.Fields[kering.Protocols.keri][kering.Vrsn_2_0][
    kering.Ilks.rpy
]
RPY_V_2, RpySchema_V_2 = dataclassFromFielddom("RPY_V_2", rpyFieldDomV2)


class RpyEscrowCollectionEnd:
    @staticmethod
    def on_get(req, rep):
        """
        GET endpoint for reply escrow collection

        Parameters:
            req (falcon.Request): The request object.
            rep (falcon.Response): The response object.

        ---
        summary: Retrieve reply escrows.
        description: This endpoint retrieves the reply escrows and can filter the collection based on a specific route.
        tags:
        - Reply Escrow
        parameters:
        - in: query
          name: route
          schema:
            type: string
          required: false
          description: The specific route to filter the reply escrow collection.
        responses:
            200:
                description: Successfully retrieved the reply escrows.
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        $ref: '#/components/schemas/Rpy'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
        """
        agent = req.context.agent

        # Optional Route parameter
        route = req.params.get("route")
        keys = (route,) if route is not None else ()
        events = []
        for saider in agent.hby.db.rpes.get(keys=keys):
            serder = agent.hby.db.rpys.get(keys=(saider.qb64,))
            events.append(serder.ked)

        rep.set_header("Content-Type", "application/json")
        rep.status = falcon.HTTP_200
        rep.data = json.dumps(events).encode("utf-8")


@dataclass
class Challenge:
    """Challenge data class"""

    words: list[str]
    dt: str = field(metadata={"marshmallow_field": fields.String(required=False)})
    said: str = field(
        default=None, metadata={"marshmallow_field": fields.String(required=False)}
    )
    authenticated: bool = field(
        default=False, metadata={"marshmallow_field": fields.Boolean(required=False)}
    )


class ChallengeCollectionEnd:
    """Resource for Challenge/Response Endpoints"""

    @staticmethod
    def on_get(req, rep):
        """Challenge GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        ---
        summary:  Get random list of words for a 2 factor auth challenge
        description:  Get the list of identifiers associated with this agent
        tags:
           - Challenge/Response
        parameters:
           - in: query
             name: strength
             schema:
                type: integer
             description:  cryptographic strength of word list
             required: false
        responses:
            200:
              description: An array of random words
              content:
                  application/json:
                    schema:
                        $ref: '#/components/schemas/Challenge'

        """
        mnem = mnemonic.Mnemonic(language="english")
        s = req.params.get("strength")
        strength = int(s) if s is not None else 128

        words = mnem.generate(strength=strength)
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        msg = dict(words=words.split(" "))
        rep.data = json.dumps(msg).encode("utf-8")


class ChallengeResourceEnd:
    """Resource for Challenge/Response Endpoints"""

    @staticmethod
    def on_post(req, rep, name):
        """Challenge POST endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name: human readable name or prefix of identifier to use to sign the challenge/response

        ---
        summary:  Sign challenge message and forward to peer identifier
        description:  Sign a challenge word list received out of bands and send `exn` peer to peer message
                      to recipient
        tags:
           - Challenge/Response
        parameters:
          - in: path
            name: name or prefix
            schema:
              type: string
            required: true
            description: Human readable alias or prefix for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Challenge response
                    properties:
                        recipient:
                          type: string
                          description: human readable alias recipient identifier to send signed challenge to
                        words:
                          type: array
                          description:  challenge in form of word list
                          items:
                              type: string
        responses:
           202:
              description: Success submission of signed challenge/response
        """
        agent = req.context.agent
        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.HTTPBadRequest(
                description="no matching Hab for alias or prefix {name}"
            )

        body = req.get_media()
        if "exn" not in body or "sig" not in body or "recipient" not in body:
            raise falcon.HTTPBadRequest(
                description="challenge response requires 'words', 'sig' and 'recipient'"
            )

        exn = body["exn"]
        sig = body["sig"]
        recpt = body["recipient"]
        serder = serdering.SerderKERI(sad=exn)

        ims = bytearray(serder.raw)
        ims.extend(sig.encode("utf-8"))

        agent.hby.psr.parseOne(ims=bytearray(ims))

        agent.exchanges.append(
            dict(said=serder.said, pre=hab.pre, rec=recpt, topic="challenge")
        )

        rep.status = falcon.HTTP_202


class ChallengeVerifyResourceEnd:
    """Resource for Challenge/Response Verification Endpoints"""

    @staticmethod
    def on_post(req, rep, source):
        """Challenge POST endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            source: qb64 AID of of source of signed response to verify

        ---
        summary:  Sign challenge message and forward to peer identifier
        description:  Sign a challenge word list received out of bands and send `exn` peer to peer message
                      to recipient
        tags:
           - Challenge/Response
        parameters:
          - in: path
            name: source
            schema:
              type: string
            required: true
            description: Human readable alias for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Challenge response
                    properties:
                        recipient:
                          type: string
                          description: human readable alias recipient identifier to send signed challenge to
                        words:
                          type: array
                          description:  challenge in form of word list
                          items:
                              type: string
        responses:
           202:
              description: Success submission of signed challenge/response
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/ChallengeOperation'
        """
        agent = req.context.agent

        body = req.get_media()
        words = httping.getRequiredParam(body, "words")
        if source not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"challenge response source={source} not found"
            )

        meta = dict(words=words)
        op = agent.monitor.submit(source, longrunning.OpTypes.challenge, metadata=meta)
        rep.status = falcon.HTTP_202
        rep.content_type = "application/json"
        rep.data = op.to_json().encode("utf-8")

        rep.status = falcon.HTTP_202

    @staticmethod
    def on_put(req, rep, source):
        """Challenge PUT accept endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            source: qb64 AID of of source of signed response to verify

        ---
        summary:  Mark challenge response exn message as signed
        description:  Mark challenge response exn message as signed
        tags:
           - Challenge/Response
        parameters:
        - in: path
          name: source
          schema:
            type: string
          required: true
          description: Human readable alias for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Challenge response
                    properties:
                        aid:
                          type: string
                          description: aid of signer of accepted challenge response
                        said:
                          type: array
                          description:  SAID of challenge message signed
                          items:
                              type: string
        responses:
           202:
              description: Success submission of signed challenge/response
        """
        agent = req.context.agent
        body = req.get_media()
        if "said" not in body:
            raise falcon.HTTPBadRequest(
                description="challenge response acceptance requires 'aid' and 'said'"
            )

        if source not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"challenge response source={source} not found"
            )

        said = body["said"]
        saider = coring.Saider(qb64=said)
        agent.hby.db.chas.add(keys=(source,), val=saider)

        rep.status = falcon.HTTP_202


@dataclass
class WellKnown:
    """Data class for Well Known URLs"""

    url: str = field(metadata={"marshmallow_field": fields.String(required=True)})
    dt: str = field(metadata={"marshmallow_field": fields.String(required=True)})


@dataclass
class MemberEnds:
    agent: Optional[Dict[str, str]] = None
    controller: Optional[Dict[str, str]] = None
    witness: Optional[Dict[str, str]] = None
    registrar: Optional[Dict[str, str]] = None
    watcher: Optional[Dict[str, str]] = None
    judge: Optional[Dict[str, str]] = None
    juror: Optional[Dict[str, str]] = None
    peer: Optional[Dict[str, str]] = None
    mailbox: Optional[Dict[str, str]] = None


@dataclass
class Contact:
    id: str = field(metadata={"marshmallow_field": fields.String(required=True)})
    alias: str = field(
        default=None, metadata={"marshmallow_field": fields.String(required=False)}
    )
    oobi: str = field(
        default=None, metadata={"marshmallow_field": fields.String(required=False)}
    )
    ends: MemberEnds = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(MemberEnds), allow_none=False
            )
        },
    )
    challenges: List[Challenge] = field(
        default=None,
        metadata={
            "marshmallow_field": fields.List(
                fields.Nested(class_schema(Challenge), allow_none=False)
            )
        },
    )
    wellKnowns: List[WellKnown] = field(
        default=None,
        metadata={
            "marshmallow_field": fields.List(
                fields.Nested(class_schema(WellKnown), allow_none=False)
            )
        },
    )

    # override this in spec to add additional fields


class ContactCollectionEnd:
    def on_get(self, req, rep):
        """Contact plural GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
        ---
        summary:  Get list of contact information associated with remote identifiers
        description:  Get list of contact information associated with remote identifiers.  All
                      information is metadata and kept in local storage only
        tags:
           - Contacts
        parameters:
          - in: query
            name: group
            schema:
              type: string
            required: false
            description: field name to group results by
          - in: query
            name: filter_field
            schema:
               type: string
            description: field name to search
            required: false
          - in: query
            name: filter_value
            schema:
               type: string
            description: value to search for
            required: false
        responses:
           200:
              description: List of contact information for remote identifiers
              content:
                application/json:
                    schema:
                        type: array
                        items:
                            $ref: '#/components/schemas/Contact'
        """
        # TODO:  Add support for sorting
        agent = req.context.agent
        group = req.params.get("group")
        field = req.params.get("filter_field")
        val = req.params.get("filter_value")

        if group is not None:
            data = dict()
            values = agent.org.values(group, val)
            for value in values:
                contacts = agent.org.find(group, value)
                self.authn(agent, contacts)
                data[value] = contacts

            rep.status = falcon.HTTP_200
            rep.data = json.dumps(data).encode("utf-8")

        elif field is not None:
            val = req.params.get("filter_value")
            if val is None:
                raise falcon.HTTPBadRequest(
                    description="filter_value if required if field_field is specified"
                )

            contacts = agent.org.find(field=field, val=val)
            self.authn(agent, contacts)
            rep.status = falcon.HTTP_200
            rep.data = json.dumps(contacts).encode("utf-8")

        else:
            data = []
            contacts = agent.org.list()

            for contact in contacts:
                aid = contact["id"]
                if aid in agent.hby.kevers and aid not in agent.hby.prefixes:
                    data.append(contact)

            self.authn(agent, data)
            rep.status = falcon.HTTP_200
            rep.data = json.dumps(data).encode("utf-8")

    @staticmethod
    def authn(agent, contacts):
        for contact in contacts:
            aid = contact["id"]

            ends = agent.agentHab.endsFor(aid)
            contact["ends"] = ends

            accepted = [saider.qb64 for saider in agent.hby.db.chas.get(keys=(aid,))]
            received = [saider.qb64 for saider in agent.hby.db.reps.get(keys=(aid,))]

            challenges = []
            for said in received:
                exn = agent.hby.db.exns.get(keys=(said,))
                challenges.append(
                    dict(
                        dt=exn.ked["dt"],
                        words=exn.ked["a"]["words"],
                        said=said,
                        authenticated=said in accepted,
                    )
                )

            contact["challenges"] = challenges

            wellKnowns = []
            wkans = agent.hby.db.wkas.get(keys=(aid,))
            for wkan in wkans:
                wellKnowns.append(dict(url=wkan.url, dt=wkan.dt))

            contact["wellKnowns"] = wellKnowns


class ContactImageResourceEnd:
    @staticmethod
    def on_post(req, rep, prefix):
        """

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier prefix of contact to associate with image

        ---
         summary: Uploads an image to associate with identifier.
         description: Uploads an image to associate with identifier.
         tags:
            - Contacts
         parameters:
           - in: path
             name: prefix
             schema:
                type: string
             description: identifier prefix to associate image to
         requestBody:
             required: true
             content:
                image/jpg:
                  schema:
                    type: string
                    format: binary
                image/png:
                  schema:
                    type: string
                    format: binary
         responses:
           200:
              description: Image successfully uploaded

        """
        agent = req.context.agent
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"{prefix} is not a known identifier."
            )

        if req.content_length > 1000000:
            raise falcon.HTTPBadRequest(description="image too big to save")

        agent.org.setImg(pre=prefix, typ=req.content_type, stream=req.bounded_stream)
        rep.status = falcon.HTTP_202

    @staticmethod
    def on_get(req, rep, prefix):
        """Contact image GET endpoint

         Parameters:
             req: falcon.Request HTTP request
             rep: falcon.Response HTTP response
             prefix: qb64 identifier prefix of contact information to get

        ---
         summary:  Get contact image for identifer prefix
         description:  Get contact image for identifer prefix
         tags:
            - Contacts
         parameters:
           - in: path
             name: prefix
             schema:
               type: string
             required: true
             description: qb64 identifier prefix of contact image to get
         responses:
            200:
               description: Contact information successfully retrieved for prefix
               content:
                   image/jpg:
                     schema:
                         description: Image
                         type: binary
            404:
               description: No contact information found for prefix
        """
        agent = req.context.agent
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"{prefix} is not a known identifier."
            )

        data = agent.org.getImgData(pre=prefix)
        if data is None:
            raise falcon.HTTPNotFound(description=f"no image available for {prefix}.")

        rep.status = falcon.HTTP_200
        rep.set_header("Content-Type", data["type"])
        rep.set_header("Content-Length", data["length"])
        rep.stream = agent.org.getImg(pre=prefix)


class ContactResourceEnd:
    @staticmethod
    def on_get(req, rep, prefix):
        """Contact GET endpoint

         Parameters:
             req: falcon.Request HTTP request
             rep: falcon.Response HTTP response
             prefix: qb64 identifier prefix of contact information to get

        ---
         summary:  Get contact information associated with single remote identifier
         description:  Get contact information associated with single remote identifier.  All
                       information is meta-data and kept in local storage only
         tags:
            - Contacts
         parameters:
           - in: path
             name: prefix
             schema:
               type: string
             required: true
             description: qb64 identifier prefix of contact to get
         responses:
            200:
               description: Contact information successfully retrieved for prefix
               content:
                 application/json:
                   schema:
                     $ref: '#/components/schemas/Contact'
            404:
               description: No contact information found for prefix
        """
        agent = req.context.agent
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"{prefix} is not a known identifier."
            )

        contact = agent.org.get(prefix)
        if contact is None:
            raise falcon.HTTPNotFound(description="NOT FOUND")

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(contact).encode("utf-8")

    @staticmethod
    def on_post(req, rep, prefix):
        """Contact plural GET endpoint

         Parameters:
             req: falcon.Request HTTP request
             rep: falcon.Response HTTP response
             prefix: human readable name of identifier to replace contact information

        ---
         summary:  Create new contact information for an identifier
         description:  Creates new information for an identifier, overwriting all existing
                       information for that identifier
         tags:
            - Contacts
         parameters:
           - in: path
             name: prefix
             schema:
               type: string
             required: true
             description: qb64 identifier prefix to add contact metadata to
         requestBody:
             required: true
             content:
               application/json:
                 schema:
                     description: Contact information
                     type: object

         responses:
            200:
               description: Updated contact information for remote identifier
               content:
                 application/json:
                   schema:
                     $ref: '#/components/schemas/Contact'
            400:
               description: Invalid identifier used to update contact information
            404:
               description: Prefix not found in identifier contact information
        """
        agent = req.context.agent
        body = req.get_media()
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description="{prefix} is not a known identifier.  oobi required before contact "
                "information"
            )

        if prefix in agent.hby.prefixes:
            raise falcon.HTTPBadRequest(
                description=f"{prefix} is a local identifier, contact information only for "
                f"remote identifiers"
            )

        if "id" in body:
            del body["id"]

        if agent.org.get(prefix):
            raise falcon.HTTPBadRequest(
                description=f"contact data for {prefix} already exists"
            )

        agent.org.replace(prefix, body)
        contact = agent.org.get(prefix)

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(contact).encode("utf-8")

    @staticmethod
    def on_put(req, rep, prefix):
        """Contact PUT endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier to update contact information

        ---
        summary:  Update provided fields in contact information associated with remote identifier prefix
        description:  Update provided fields in contact information associated with remote identifier prefix.  All
                      information is metadata and kept in local storage only
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix to add contact metadata to
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Contact information
                    type: object

        responses:
           200:
              description: Updated contact information for remote identifier
              content:
                  application/json:
                     schema:
                        $ref: '#/components/schemas/Contact'
           400:
              description: Invalid identifier used to update contact information
           404:
              description: Prefix not found in identifier contact information
        """
        agent = req.context.agent
        body = req.get_media()
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"{prefix} is not a known identifier.  oobi required before contact information"
            )

        if prefix in agent.hby.prefixes:
            raise falcon.HTTPBadRequest(
                description=f"{prefix} is a local identifier, contact information only for remote identifiers"
            )

        if "id" in body:
            del body["id"]

        agent.org.update(prefix, body)
        contact = agent.org.get(prefix)

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(contact).encode("utf-8")

    @staticmethod
    def on_delete(req, rep, prefix):
        """Contact plural GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier prefix to delete contact information

        ---
        summary:  Delete contact information associated with remote identifier
        description:  Delete contact information associated with remote identifier
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix of contact to delete
        responses:
           202:
              description: Contact information successfully deleted for prefix
           404:
              description: No contact information found for prefix
        """
        agent = req.context.agent
        deleted = agent.org.rem(prefix)
        if not deleted:
            raise falcon.HTTPNotFound(
                description=f"no contact information to delete for {prefix}"
            )

        rep.status = falcon.HTTP_202


@dataclass
class AidRecord:
    aid: str
    ends: MemberEnds


@dataclass
class GroupMember:
    signing: List[AidRecord]
    rotation: List[AidRecord]


class GroupMemberCollectionEnd:
    @staticmethod
    def on_get(req, rep, name):
        """
        GET endpoint for group members
        Parameters:
            req (falcon.Request): The request object.
            rep (falcon.Response): The response object.
            name (str): The human-readable name or prefix of the identifier.

        ---
        summary: Fetch group member information.
        description: This endpoint retrieves the signing and rotation members for a specific group associated with an identifier.
        tags:
        - Group Member
        parameters:
        - in: path
          name: name or prefix
          schema:
            type: string
          required: true
          description: The human-readable name of the identifier or its prefix.
        responses:
            200:
                description: Successfully fetched the group member information.
                content:
                  application/json:
                    schema:
                      $ref: '#/components/schemas/GroupMember'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
            404:
                description: The requested identifier was not found.
        """
        agent = req.context.agent

        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.errors.HTTPNotFound(
                description=f"invalid alias or prefix {name}"
            )

        if not isinstance(hab, habbing.SignifyGroupHab):
            raise falcon.HTTPBadRequest(
                description="members endpoint only available for group AIDs"
            )

        smids = hab.db.signingMembers(hab.pre)
        rmids = hab.db.rotationMembers(hab.pre)

        signing = []
        for smid in smids:
            ends = hab.endsFor(smid)
            signing.append(dict(aid=smid, ends=ends))

        rotation = []
        for rmid in rmids:
            ends = hab.endsFor(rmid)
            rotation.append(dict(aid=rmid, ends=ends))

        data = dict(signing=signing, rotation=rotation)
        rep.status = falcon.HTTP_200
        rep.data = json.dumps(data).encode("utf-8")



# -*- encoding: utf-8 -*-
"""
main package
"""

__version__ = "0.4.0"  # also change in setup.py

import logging
from hio.help import ogling
from .monitoring.logs import TruncatedFormatter

log_name = "keria"  # name of this project that shows up in log messages
log_format_str = f"%(asctime)s [{log_name}] %(levelname)-8s %(module)s.%(funcName)s-%(lineno)s %(message)s"

ogler = ogling.initOgler(prefix=log_name, syslogged=False)
keriLogger = ogler.getLogger()
ogler.level = logging.INFO

formatter = TruncatedFormatter(log_format_str)
formatter.default_msec_format = None
logHandler = logging.StreamHandler()
logHandler.setFormatter(formatter)
ogler.baseFormatter = formatter
ogler.baseConsoleHandler = logHandler
ogler.baseConsoleHandler.setFormatter(formatter)
ogler.reopen(name=log_name, temp=True, clear=True)


def set_log_level(loglevel, logger):
    """Set the log level for the logger."""
    ogler.level = logging.getLevelName(loglevel.upper())
    keriLogger.setLevel(ogler.level)
    logger.setLevel(ogler.level)



# -*- encoding: utf-8 -*-
"""
KERIA
keria.core.longrunning module

"""

import datetime
from collections import namedtuple
from dataclasses import dataclass, asdict, field
from marshmallow import fields
from typing import Literal, Optional, Dict, Union

import falcon
import json
from dataclasses_json import dataclass_json
from keri import kering
from keri.app.oobiing import Result
from keri.core import eventing, coring, serdering
from keri.db import dbing, koming
from keri.help import helping

from keria.app import delegating

# long running operation types
Typeage = namedtuple(
    "Tierage",
    "oobi witness delegation group query registry credential endrole "  # type: ignore[name-match]
    "locscheme challenge exchange submit done",
)

OpTypes = Typeage(
    oobi="oobi",
    witness="witness",
    delegation="delegation",
    group="group",
    query="query",
    registry="registry",
    credential="credential",
    endrole="endrole",
    locscheme="locscheme",
    challenge="challenge",
    exchange="exchange",
    submit="submit",
    done="done",
)


@dataclass_json
@dataclass
class OperationStatus:
    code: int
    message: str
    details: Optional[Dict] = field(
        default=None, metadata={"marshmallow_field": fields.Dict(allow_none=True)}
    )


@dataclass_json
@dataclass
class BaseOperation:
    name: str
    metadata: Optional[dict]


@dataclass_json
@dataclass
class PendingOperation(BaseOperation):
    done: Literal[False] = False


@dataclass_json
@dataclass
class CompletedOperation(BaseOperation):
    response: dict
    done: Literal[True] = True


@dataclass_json
@dataclass
class FailedOperation(BaseOperation):
    error: OperationStatus
    done: Literal[True] = True


Operation = Union[PendingOperation, CompletedOperation, FailedOperation]


@dataclass
class Op:
    oid: str
    type: str
    start: str
    metadata: dict


class Operator(dbing.LMDBer):
    TailDirPath = "keri/opr"
    AltTailDirPath = ".keri/opr"
    TempPrefix = "keri_ops_"

    def __init__(self, name="opr", headDirPath=None, reopen=True, **kwa):
        """

        Parameters:
            headDirPath:
            perm:
            reopen:
            kwa:
        """
        self.ops = None
        self.msgs = None

        super(Operator, self).__init__(
            name=name, headDirPath=headDirPath, reopen=reopen, **kwa
        )

    def reopen(self, **kwa):
        """Reopen database and initialize sub-dbs"""
        super(Operator, self).reopen(**kwa)

        # Long running operations, keyed by "name" which is f"{type}.{oid}"
        self.ops = koming.Komer(
            db=self,
            subkey="opr.",
            schema=Op,
        )

        return self.env


class Monitor:
    """Monitoring and garbage collecting long running operations

    Attributes:
        hby (Habery): identifier database environment
        opr(Operator): long running operations database
        swain(Anchorer): Delegation processes tracker

    """

    def __init__(
        self,
        hby,
        swain,
        counselor=None,
        registrar=None,
        exchanger=None,
        credentialer=None,
        submitter=None,
        opr=None,
        temp=False,
    ):
        """Create long running operation monitor

        Parameters:
            hby (Habery): identifier database environment
            swain(Anchorer): Delegation processes tracker
            opr (Operator): long running operations database

        """
        self.hby = hby
        self.swain = swain
        self.counselor = counselor
        self.registrar = registrar
        self.exchanger = exchanger
        self.credentialer = credentialer
        self.submitter = submitter
        self.opr = opr if opr is not None else Operator(name=hby.name, temp=temp)

    def submit(self, oid, typ, metadata=None):
        """Submit a new long running operation to track

        parameters:
            oid (str): opaque identifier of the target of the long running operation (type specific)
            typ (Typeage): long running operation type
            metadata (dict): additional metadata for the long running operation (type specific)

        Returns:
            Operation: the initial long running operation status object

        """
        if typ not in OpTypes:
            raise kering.InvalidValueError(
                f"{typ} not a valid long running operation type"
            )

        name = f"{typ}.{oid}"
        op = Op(oid=oid, type=typ, start=helping.nowIso8601(), metadata=metadata)

        # Overwrite any existing long running operation of this type for this resource.
        # resets the clock basically
        self.opr.ops.pin(keys=(name,), val=op)

        # Return Operation with full status check in case its already finished.
        return self.get(name)

    def get(self, name):
        if (op := self.opr.ops.get(keys=(name,))) is None:
            return None

        operation = self.status(op)

        return operation

    def getOperations(self, type=None):
        """Return list of long running opterations, optionally filtered by type"""
        ops = self.opr.ops.getItemIter()
        if type is not None:
            ops = filter(lambda i: i[1].type == type, ops)

        def get_status(op):
            try:
                return self.status(op)
            except Exception as err:
                # self.status may throw an exception.
                # Handling error by returning an operation with error status
                return FailedOperation(
                    name=f"{op.type}.{op.oid}",
                    metadata=op.metadata,
                    done=True,
                    error=OperationStatus(code=500, message=f"{err}"),
                )

        return [get_status(op) for (_, op) in ops]

    def rem(self, name):
        """Remove tracking of the long running operation represented by name"""
        return self.opr.ops.rem(keys=(name,))

    def status(self, op):
        """Calculate the status of an operation.

        Base on the type of an operation, determine the current status of the operation, including loading
        any availabie error messages if the operation failed of the sucessful result of the operation which
        will be the same that would be returned by the original endpoint if the operation were not long running

        Parameters:
            op (Op): database storage for long running operation

        Returns:
            Operation: The status of the operation, including any errors or results as appropriate.

        """

        done = False
        response = None
        error = None
        name = f"{op.type}.{op.oid}"
        metadata = op.metadata

        if op.type in (OpTypes.witness,):
            if "pre" not in op.metadata or "sn" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required fields ('pre', 'sn')"
                )

            pre = op.metadata["pre"]
            if pre not in self.hby.kevers:
                raise kering.ValidationError(
                    f"identifier {pre} not found for long running {op.oid} ({op.type})"
                )

            sn = op.metadata["sn"]
            kever = self.hby.kevers[pre]
            sdig = self.hby.db.getKeLast(
                key=dbing.snKey(pre=kever.prefixer.qb64b, sn=sn)
            )
            if sdig is not None:
                dgkey = dbing.dgKey(kever.prefixer.qb64b, bytes(sdig))
                wigs = self.hby.db.getWigs(dgkey)

                if len(wigs) >= kever.toader.num:
                    evt = self.hby.db.getEvt(
                        dbing.dgKey(pre=kever.prefixer.qb64, dig=bytes(sdig))
                    )
                    serder = serdering.SerderKERI(raw=bytes(evt))
                    done = True
                    response = serder.ked

                else:
                    start = helping.fromIso8601(op.start)
                    dtnow = helping.nowUTC()
                    if (dtnow - start) > datetime.timedelta(
                        seconds=eventing.Kevery.TimeoutPWE
                    ):
                        done = True
                        error = OperationStatus(
                            code=408,  # Using HTTP error codes here for lack of a better alternative
                            message=f"long running {op.type} for {op.oid} (pre: {pre}) operation timed out before "
                            f"receiving sufficient witness receipts",
                        )
                    else:
                        done = False

            else:
                done = False

        elif op.type in (OpTypes.oobi,):
            if "oobi" not in op.metadata:
                raise kering.ValidationError(
                    "invalid OOBI long running operation, missing oobi"
                )

            oobi = op.metadata["oobi"]
            obr = self.hby.db.roobi.get(keys=(oobi,))
            if obr is None:
                done = False
            elif obr.state == Result.resolved:
                done = True
                if obr.cid and obr.cid in self.hby.kevers:
                    kever = self.hby.kevers[obr.cid]
                    response = asdict(kever.state())
                else:
                    response = dict(oobi=oobi)

            elif obr.state == Result.failed:
                done = True
                error = OperationStatus(
                    code=500, message=f"resolving OOBI {op.oid} failed"
                )
            else:
                done = False

        elif op.type in (OpTypes.delegation,):
            if "pre" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required field 'pre'"
                )

            pre = op.metadata["pre"]
            if pre not in self.hby.kevers:
                raise kering.ValidationError(
                    f"long running {op.type} operation identifier {pre} not found (oid: {op.oid})"
                )

            kever = self.hby.kevers[pre]

            reqsn = "sn"
            reqtee = "teepre"
            anchor = "anchor"
            required = [reqsn, reqtee]
            if reqsn in op.metadata:  # delegatee detects successful delegation
                sn = op.metadata["sn"]
                seqner = coring.Seqner(sn=sn)
                sdig = self.hby.db.getKeLast(key=dbing.snKey(pre=pre, sn=sn))

                if self.swain.complete(kever.prefixer, seqner):
                    evt = self.hby.db.getEvt(
                        dbing.dgKey(pre=kever.prefixer.qb64, dig=bytes(sdig))
                    )
                    serder = serdering.SerderKERI(raw=bytes(evt))

                    done = True
                    response = serder.ked
                else:
                    done = False
            elif (
                reqtee in op.metadata
            ):  # delegator detects delegatee delegation success
                teepre = op.metadata[reqtee]
                anc = op.metadata[anchor]
                if (
                    teepre in self.hby.kevers
                ):  # delegatee dip has been processed by the delegator
                    done = True
                    response = op.metadata[reqtee]
                else:
                    hab = self.hby.habByPre(kever.prefixer.qb64)
                    delegating.approveDelegation(hab, anc)
                    done = False
            else:
                raise falcon.HTTPBadRequest(
                    description=f"longrunning operation type {op.type} requires one of {required}, but are missing from request"
                )

        elif op.type in (OpTypes.group,):
            if "pre" not in op.metadata or "sn" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required fields ('pre', 'sn')"
                )

            pre = op.metadata["pre"]
            prefixer = coring.Prefixer(qb64=pre)
            seqner = coring.Seqner(sn=op.metadata["sn"])

            if self.counselor.complete(prefixer, seqner):
                sdig = self.hby.db.getKeLast(key=dbing.snKey(pre=pre, sn=seqner.sn))
                evt = self.hby.db.getEvt(
                    dbing.dgKey(pre=prefixer.qb64, dig=bytes(sdig))
                )
                serder = serdering.SerderKERI(raw=bytes(evt))

                done = True
                response = serder.ked
            else:
                done = False

        elif op.type in (OpTypes.query,):
            if "pre" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required field 'pre'"
                )

            pre = op.metadata["pre"]
            if pre not in self.hby.kevers:
                done = False

            else:
                kever = self.hby.kevers[pre]
                if "sn" in op.metadata:
                    sn = int(op.metadata["sn"], 16)
                    if kever.sn >= sn:
                        done = True
                        response = asdict(kever.state())
                    else:
                        done = False
                elif "anchor" in op.metadata:
                    anchor = op.metadata["anchor"]
                    if self.hby.db.findAnchoringSealEvent(pre, seal=anchor) is not None:
                        done = True
                        response = asdict(kever.state())
                    else:
                        done = False
                else:
                    ksn = None
                    for _, saider in self.hby.db.knas.getItemIter(keys=(pre,)):
                        ksn = self.hby.db.ksns.get(keys=(saider.qb64,))
                        break

                    if ksn and ksn.d == kever.serder.said:
                        done = True
                        response = asdict(kever.state())
                    else:
                        done = False

        elif op.type in (OpTypes.registry,):
            if "pre" not in op.metadata or "anchor" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required fields ('pre', 'anchor')"
                )

            pre = op.metadata["pre"]
            if pre not in self.hby.kevers:
                raise kering.ValidationError(
                    f"identifier {pre} for long running {op.type} operation {op.oid} not found"
                )

            anchor = op.metadata["anchor"]
            if self.hby.db.findAnchoringSealEvent(pre, seal=anchor) is not None:
                done = True
                response = dict(anchor=anchor)
            else:
                done = False

        elif op.type in (OpTypes.credential,):
            if "ced" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'ced' field"
                )

            ced = op.metadata["ced"]
            if self.credentialer.complete(ced["d"]):
                done = True
                response = dict(ced=ced)
            else:
                done = False

        elif op.type in (OpTypes.exchange,):
            if self.exchanger.complete(op.oid):
                done = True
                response = dict(said=op.oid)
            else:
                done = False

        elif op.type in (OpTypes.endrole,):
            if (
                "cid" not in op.metadata
                or "role" not in op.metadata
                or "eid" not in op.metadata
            ):
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required fields ('cid', 'role', 'eid')"
                )

            cid = op.metadata["cid"]
            role = op.metadata["role"]
            eid = op.metadata["eid"]

            end = self.hby.db.ends.get(keys=(cid, role, eid))
            if end and (end.enabled or end.allowed):
                saider = self.hby.db.eans.get(keys=(cid, role, eid))
                serder = self.hby.db.rpys.get(keys=(saider.qb64,))
                done = True
                response = serder.ked
            else:
                done = False

        elif op.type in (OpTypes.locscheme,):
            if (
                "eid" not in op.metadata
                or "scheme" not in op.metadata
                or "url" not in op.metadata
            ):
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing required fields ('eid', 'scheme', 'url')"
                )

            eid = op.metadata["eid"]
            scheme = op.metadata["scheme"]
            url = op.metadata["url"]

            loc = self.hby.db.locs.get(keys=(eid, scheme))
            if loc:
                done = True
                response = dict(eid=eid, scheme=scheme, url=url)
            else:
                done = False

        elif op.type in (OpTypes.challenge,):
            if op.oid not in self.hby.kevers:
                done = False

            if "words" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'words' field"
                )

            found = False
            words = op.metadata["words"]
            saiders = self.hby.db.reps.get(keys=(op.oid,))
            for saider in saiders:
                exn = self.hby.db.exns.get(keys=(saider.qb64,))
                if words == exn.ked["a"]["words"]:
                    found = True
                    break

            if found:
                done = True
                response = dict(exn=exn.ked)
            else:
                done = False

        elif op.type in (OpTypes.submit,):
            kever = self.hby.kevers[op.oid]
            if (
                kever
                and len(self.submitter.submits) == 0
                and len(self.submitter.doers) == 0
            ):
                done = True
                response = asdict(kever.state())
            else:
                start = helping.fromIso8601(op.start)
                dtnow = helping.nowUTC()
                if (dtnow - start) > datetime.timedelta(
                    seconds=eventing.Kevery.TimeoutPWE
                ):
                    done = True
                    error = OperationStatus(
                        code=408,  # Using HTTP error codes here for lack of a better alternative
                        message=f"long running {op.type} for {op.oid} operation timed out before "
                        f"receiving sufficient witness receipts",
                    )
                else:
                    done = False

        elif op.type in (OpTypes.done,):
            done = True
            response = op.metadata["response"]

        else:
            done = True
            error = OperationStatus(
                code=404,  # Using HTTP error codes here for lack of a better alternative
                message=f"long running operation type {op.type} unknown",
            )

        if error is not None:
            return FailedOperation(
                name=name,
                metadata=metadata,
                error=error,
            )
        elif done:
            return CompletedOperation(
                name=name,
                metadata=metadata,
                response=response,
            )
        else:
            return PendingOperation(
                name=name,
                metadata=metadata,
            )


class OperationCollectionEnd:
    @staticmethod
    def on_get(req, rep):
        """Get list of long running operations

        Parameters:
            req (Request):  Falcon HTTP Request object
            rep (Response): Falcon HTTP Response object

        ---
        summary: Get list of long running operations
        parameters:
          - in: query
            name: type
            schema:
              type: string
            required: false
            description: filter list of long running operations by type
        responses:
            200:
              description: list of long running operations
              content:
                  application/json:
                    schema:
                        type: array
                        items:
                          $ref: '#/components/schemas/Operation'
        """
        agent = req.context.agent
        type = req.params.get("type")
        ops = agent.monitor.getOperations(type=type)
        rep.data = json.dumps(ops, default=lambda o: o.to_dict()).encode("utf-8")
        rep.content_type = "application/json"
        rep.status = falcon.HTTP_200


class OperationResourceEnd:
    """Single Resource REST endpoint for long running operations"""

    @staticmethod
    def on_get(req, rep, name):
        """GET single resource REST endpoint

        Parameters:
            req (Request):  Falcon HTTP Request object
            rep (Response): Falcon HTTP Response object
            name (str): Long running operation resource name to load
        ---
        summary: Retrieve a specific long running operation.
        description: This endpoint retrieves the status of a long running operation by its name.
        tags:
        - Operation
        parameters:
          - in: path
            name: name
            schema:
              type: string
            required: true
            description: The name of the long running operation to retrieve.
        responses:
          200:
            description: Successfully retrieved the status of the long running operation.
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/Operation'
          404:
            description: The requested long running operation was not found.

        """
        agent = req.context.agent
        if (operation := agent.monitor.get(name)) is None:
            raise falcon.HTTPNotFound(
                title=f"long running operation '{name}' not found"
            )

        rep.content_type = "application/json"
        rep.data = operation.to_json().encode("utf-8")
        rep.status = falcon.HTTP_200

    @staticmethod
    def on_delete(req, rep, name):
        """DELETE single resource REST endpoint

        Args:
            req (Request):  Falcon HTTP Request object
            rep (Response): Falcon HTTP Response object
            name (str): Long running operation resource name to load
        ---
        summary: Remove a specific long running operation.
        description: This endpoint removes a long running operation by its name.
        tags:
        - Operation
        parameters:
          - in: path
            name: name
            schema:
              type: string
            required: true
            description: The name of the long running operation to remove.
        responses:
          204:
              description: Successfully removed the long running operation.
          404:
              description: The requested long running operation was not found.
          500:
              description: Internal server error. This could be due to an issue with removing the operation.
        """

        agent = req.context.agent
        if agent.monitor.get(name) is None:
            raise falcon.HTTPNotFound(
                title=f"long running operation '{name}' not found"
            )

        deleted = agent.monitor.rem(name)
        if deleted:
            rep.status = falcon.HTTP_204
        else:
            raise falcon.HTTPInternalServerError(
                f"unable to delete long running operation {name}"
            )



# -*- encoding: utf-8 -*-
"""
KERIA
keria.core.httping module

"""

import io
import logging

import falcon
from falcon.http_status import HTTPStatus
from hio.core import tcp, http

from keria import ogler, log_name

logger = ogler.getLogger(log_name)


class HandleCORS(object):
    def process_request(self, req, resp):
        resp.set_header("Access-Control-Allow-Origin", "*")
        resp.set_header("Access-Control-Allow-Methods", "*")
        resp.set_header("Access-Control-Allow-Headers", "*")
        resp.set_header("Access-Control-Max-Age", 1728000)  # 20 days
        if req.method == "OPTIONS":
            raise HTTPStatus(falcon.HTTP_200)


def getRequiredParam(body, name):
    param = body.get(name)
    if param is None:
        raise falcon.HTTPBadRequest(
            description=f"required field '{name}' missing from request"
        )

    return param


def parseRangeHeader(header, name, start=0, end=9):
    """Parse the start and end requested range values, defaults are 0, 9

    Parameters:
        header(str):  HTTP Range header value
        name (str): range name to look for
        start (int): default start index
        end(int): default end index

    Returns:
        (start, end): tuple of start index and end index

    """

    if not header.startswith(f"{name}="):
        return start, end

    header = header.strip(f"{name}=")
    try:
        if header.startswith("-"):
            return start, int(header[1:])

        if header.endswith("-"):
            return int(header[:-1]), end

        vals = header.split("-")
        if not len(vals) == 2:
            return start, end

        return int(vals[0]), int(vals[1])
    except ValueError:
        return start, end


# noinspection PyMethodMayBeStatic
class RequestLoggerMiddleware:
    """Simple request logger Falcon middleware."""

    def process_request(self, req: falcon.Request, resp: falcon.Response):
        """Log incoming requests."""
        logger.info("Request received : %s %s", req.method, req.url)
        logger.debug("Request headers : %s", req.headers)
        if req.content_length and logger.isEnabledFor(logging.DEBUG):
            # Read and re-set the stream to allow further processing
            body = req.stream.read()
            decoded_body = body.decode("utf-8") if body else "<empty>"
            logger.debug("Request body    : %s", decoded_body)
            req.env["wsgi.input"] = io.BytesIO(
                body
            )  # Reset the stream for further processing
            req.env["CONTENT_LENGTH"] = str(
                len(body)
            )  # match WSGI env content length to body length
        else:
            logger.debug("Request body    : No body")

    def process_response(
        self, req: falcon.Request, resp: falcon.Response, resource, req_succeeded
    ):
        """Log outgoing responses."""
        logger.info("Response status  : %s on %s %s", resp.status, req.method, req.url)
        logger.debug("Response headers: %s", resp.headers)


keriHeaders = [
    "cesr-attachment",
    "cesr-date",
    "content-type",
    "signature",
    "signature-input",
    "signify-resource",
    "signify-timestamp",
]

corsMiddleware = falcon.CORSMiddleware(
    allow_origins="*", allow_credentials="*", expose_headers=keriHeaders
)


def falconApp(logRequests=False):
    middlewares: list = [corsMiddleware]
    if logRequests:
        middlewares.append(RequestLoggerMiddleware())
    return falcon.App(middleware=middlewares)


def createHttpServer(port, app, keypath=None, certpath=None, cafilepath=None):
    """
    Create an HTTP or HTTPS server depending on whether TLS key material is present

    Parameters:
        port (int)         : port to listen on for all HTTP(s) server instances
        app (falcon.App)   : application instance to pass to the http.Server instance
        keypath (string)   : the file path to the TLS private key
        certpath (string)  : the file path to the TLS signed certificate (public key)
        cafilepath (string): the file path to the TLS CA certificate chain file
    Returns:
        hio.core.http.Server
    """
    if keypath is not None and certpath is not None and cafilepath is not None:
        servant = tcp.ServerTls(
            certify=False,
            keypath=keypath,
            certpath=certpath,
            cafilepath=cafilepath,
            port=port,
        )
        server = http.Server(port=port, app=app, servant=servant)
    else:
        server = http.Server(port=port, app=app)
    return server



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""

import argparse

from hio.base import doing
from keri import help, kering
from keri.app import habbing
from keri.app.cli.common import existing
from keri.core import serdering, coring
from keri.db import basing, dbing
from keria.db import basing as abase

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(
    description="Migrates existing public key/next key digest data from First Seen Event logs to signing member IDs "
    "(smids) and rotation member IDs (rmids) on SignifyGroupHabs"
)
parser.set_defaults(handler=lambda args: handler(args), transferable=True)
parser.add_argument(
    "--base",
    "-b",
    help="additional optional prefix to file location of KERI keystore",
    required=False,
    default="",
)
parser.add_argument(
    "--force", action="store_true", required=False, default=False, help="Perform update"
)


def handler(args):
    kwa = dict(args=args)
    return [doing.doify(fix, **kwa)]


def fix(tymth, tock=0.0, **opts):
    _ = yield tock
    args = opts["args"]

    prefix_by_public_key = dict()
    prefix_by_next_key_digest = dict()

    adb = abase.AgencyBaser(name="TheAgency", base=args.base, reopen=True, temp=False)

    caids = []
    for (caid,), _ in adb.agnt.getItemIter():
        caids.append(caid)

    signify_group_habs = dict()
    for caid in caids:
        with existing.existingHby(name=caid, base=args.base) as hby:
            for pre, hab in hby.habs.items():
                if type(hab) is habbing.SignifyGroupHab:
                    signify_group_habs[pre] = hab.name

    # create caches of existing public keys and next key digests and the associated prefixes
    for caid in caids:
        db = basing.Baser(name=caid, base=args.base, temp=False, reopen=False)
        try:
            db.reopen()
        except kering.DatabaseError:
            return -1

        for pre, fn, dig in db.getFelItemAllPreIter(key=b""):
            dgkey = dbing.dgKey(pre, dig)
            if not (raw := db.getEvt(key=dgkey)):
                raise kering.MissingEntryError("Missing event for dig={}.".format(dig))

            serder = serdering.SerderKERI(raw=bytes(raw))
            val = (coring.Prefixer(qb64b=serder.preb), coring.Seqner(sn=serder.sn))
            verfers = serder.verfers or []
            ndigers = serder.ndigers or []

            if val[0].qb64 in signify_group_habs:
                continue

            for verfer in verfers:
                prefix_by_public_key[verfer.qb64] = val

            for diger in ndigers:
                prefix_by_next_key_digest[diger.qb64] = val

    # pretty
    pre_name_cache = dict()
    for caid in caids:
        with existing.existingHby(name=caid, base=args.base) as hby:
            for pre, hab in hby.habs.items():
                pre_name_cache[pre] = hab.name

    # Arby's
    # update existing hab records with the correct smids and rmids
    for caid in caids:
        with existing.existingHby(name=caid, base=args.base) as hby:
            print()
            print(f"Hby {hby.name}")

            for pre, hab in hby.habs.items():
                if type(hab) is not habbing.SignifyGroupHab:
                    print(f"\t skipping {pre} - {type(hab)}")
                    continue

                print()
                print(f"\t {hab.name} - {pre} - {type(hab)}")
                print()

                if (
                    not hasattr(hab, "smids")
                    and not hasattr(hab, "rmids")
                    or hab.smids is None
                    and hab.rmids is None
                ):
                    print()
                    smids = set()
                    rmids = set()
                    for v in hab.kever.verfers:
                        if v.qb64 in prefix_by_public_key:
                            smids.add(prefix_by_public_key[v.qb64][0].qb64)

                    for v in hab.kever.ndigers:
                        if v.qb64 in prefix_by_next_key_digest:
                            rmids.add(prefix_by_next_key_digest[v.qb64][0].qb64)

                    print(
                        f"\t Proposed smids and rmids updates for {hab.name} - {pre}:"
                    )
                    print(f"\t\t smids: {smids}")
                    for smid in smids:
                        print(f"\t\t\t -> {smid} {pre_name_cache.get(smid)}")
                    print(f"\t\t rmids: {rmids}")
                    for rmid in rmids:
                        print(f"\t\t\t -> {rmid} {pre_name_cache.get(rmid)}")

                    if args.force:
                        habr = hab.db.habs.get(keys=(hab.pre,))
                        habr.smids = list(smids)
                        habr.rmids = list(rmids)
                        print(f"\t Updating {habr}")
                        hab.db.habs.pin(keys=(hab.pre,), val=habr)
                        print()
                    else:
                        print()
                        print("no updates performed, use --force to apply changes")
                        print()



# -*- encoding: utf-8 -*-
"""
KERIA
keria.cli.keria.commands.start module

KERIA Agent server start command line interface (CLI) command
"""

import argparse
import os


from keri import __version__
from keri import help

from keria.app import agenting

d = "Runs KERI Signify Agent\n"
d += "\tExample:\nkli ahab\n"
parser = argparse.ArgumentParser(description=d)
parser.set_defaults(handler=lambda args: launch(args))
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=__version__,
    help="Prints out version of script runner.",
)
parser.add_argument(
    "-a",
    "--admin-http-port",
    dest="admin",
    action="store",
    type=int,
    default=os.getenv("KERIA_ADMIN_PORT", "3901"),
    help="Admin port number the HTTP server listens on. Default is 3901.",
)
parser.add_argument(
    "-H",
    "--http",
    action="store",
    type=int,
    default=os.getenv("KERIA_HTTP_PORT", "3902"),
    help="Local port number the HTTP server listens on. Default is 3902.",
)
parser.add_argument(
    "-B",
    "--boot",
    action="store",
    type=int,
    default=os.getenv("KERIA_BOOT_PORT", "3903"),
    help="Boot port number the Boot HTTP server listens on.  This port needs to be secured."
    " Default is 3903.",
)
parser.add_argument(
    "-n",
    "--name",
    action="store",
    default="keria",
    help="Name of controller. Default is 'keria'.",
)
parser.add_argument(
    "--base",
    "-b",
    help="additional optional prefix to file location of KERI keystore",
    required=False,
    default="",
)
parser.add_argument(
    "--passcode",
    "-p",
    help="22 character encryption passcode for keystore (is not saved)",
    dest="bran",
    default=os.getenv("KERIA_PASSCODE"),
)  # passcode => bran
parser.add_argument(
    "--config-file", dest="configFile", action="store", help="configuration filename"
)
parser.add_argument(
    "--config-dir",
    dest="configDir",
    action="store",
    default=None,
    help="directory override for configuration data",
)
parser.add_argument(
    "--keypath",
    action="store",
    required=False,
    default=None,
    help="TLS server private key file",
)
parser.add_argument(
    "--certpath",
    action="store",
    required=False,
    default=None,
    help="TLS server signed certificate (public key) file",
)
parser.add_argument(
    "--cafilepath",
    action="store",
    required=False,
    default=None,
    help="TLS server CA certificate chain",
)
parser.add_argument(
    "--loglevel",
    action="store",
    required=False,
    default=os.getenv("KERIA_LOG_LEVEL", "CRITICAL"),
    help="Set log level to DEBUG | INFO | WARNING | ERROR | CRITICAL. Default is CRITICAL",
)
parser.add_argument(
    "--logfile",
    action="store",
    required=False,
    default=None,
    help="path of the log file. If not defined, logs will not be written to the file.",
)
parser.add_argument("--logrequests", help="log HTTP requests", action="store_true")
parser.add_argument(
    "--experimental-boot-password",
    help="Experimental password for boot endpoint. Enables HTTP Basic Authentication for the boot endpoint. Only meant to be used for testing purposes.",
    dest="bootPassword",
    default=os.getenv("KERIA_EXPERIMENTAL_BOOT_PASSWORD"),
)
parser.add_argument(
    "--experimental-boot-username",
    help="Experimental username for boot endpoint. Enables HTTP Basic Authentication for the boot endpoint. Only meant to be used for testing purposes.",
    dest="bootUsername",
    default=os.getenv("KERIA_EXPERIMENTAL_BOOT_USERNAME"),
)


logger = help.ogler.getLogger()


def launch(args):
    agenting.runAgency(
        agenting.KERIAServerConfig(
            name=args.name or "ahab",
            base=args.base or "",
            bran=args.bran,
            adminPort=args.admin,
            httpPort=args.http,
            bootPort=args.boot,
            configFile=args.configFile,
            configDir=args.configDir,
            keyPath=args.keypath,
            certPath=args.certpath,
            caFilePath=args.cafilepath,
            logLevel=args.loglevel,
            logFile=args.logfile,
            logRequests=args.logrequests if args.logrequests else False,
            cors=os.getenv("KERI_AGENT_CORS", "false").lower() in ("true", "1"),
            releaseTimeout=int(os.getenv("KERIA_RELEASER_TIMEOUT", "86400")),
            curls=getListVariable("KERIA_CURLS"),
            iurls=getListVariable("KERIA_IURLS"),
            durls=getListVariable("KERIA_DURLS"),
            bootPassword=args.bootPassword,
            bootUsername=args.bootUsername,
        )
    )
    logger.info("Agent %s gracefully stopped", args.name)


def getListVariable(name):
    value = os.getenv(name)
    return value.split(";") if value else None



# -*- coding: utf-8 -*-
"""
Helpers for tests that need KERIA
"""

import json
import os
import pathlib
import shutil
from contextlib import contextmanager
from typing import Tuple, List

import falcon
from falcon import testing
from hio.core import http
from keri import kering
from keri import core
from keri.app import keeping, habbing, configing, signing
from keri.core import coring, eventing, parsing, routing, scheming, serdering
from keri.core.coring import MtrDex
from keri.core.eventing import SealEvent
from keri.core.signing import Salter, Signer
from keri.help import helping
from keri.kering import TraitCodex
from keri.vc import proving
from keri.vdr import credentialing, verifying
from keri.vdr import eventing as veventing
from keri.vdr.credentialing import Regery, Registrar

from keria.app import agenting, indirecting

SCRIPTS_DIR = str(
    pathlib.Path(
        str(pathlib.Path(__file__).parent.resolve()), "../../../tests/scripts"
    ).resolve()
)

WitnessUrls = {
    "wan:tcp": "tcp://127.0.0.1:5632/",
    "wan:http": "http://127.0.0.1:5642/",
    "wes:tcp": "tcp://127.0.0.1:5634/",
    "wes:http": "http://127.0.0.1:5644/",
    "wil:tcp": "tcp://127.0.0.1:5633/",
    "wil:http": "http://127.0.0.1:5643/",
}


class DbSeed:
    @staticmethod
    def seedWitEnds(db, witHabs, protocols=None):
        """Add endpoint and location records for well known test witnesses

        Args:
            db (Baser): database to add records
            witHabs (list): list of witness Habs for whom to create Ends
            protocols (list) array of str protocol names to load URLs for.
        Returns:

        """

        rtr = routing.Router()
        rvy = routing.Revery(db=db, rtr=rtr)
        kvy = eventing.Kevery(db=db, lax=False, local=True, rvy=rvy)
        kvy.registerReplyRoutes(router=rtr)
        psr = parsing.Parser(framed=True, kvy=kvy, rvy=rvy)

        if protocols is None:
            protocols = [kering.Schemes.tcp, kering.Schemes.http]

        for scheme in protocols:
            msgs = bytearray()
            for hab in witHabs:
                url = WitnessUrls[f"{hab.name}:{scheme}"]
                msgs.extend(
                    hab.makeEndRole(
                        eid=hab.pre,
                        role=kering.Roles.controller,
                        stamp=helping.nowIso8601(),
                    )
                )

                msgs.extend(
                    hab.makeLocScheme(
                        url=url, scheme=scheme, stamp=helping.nowIso8601()
                    )
                )
                psr.parse(ims=msgs)

    @staticmethod
    def seedSchema(db):
        sad = {
            "$id": "EBfdlu8R27Fbx-ehrqwImnK-8Cm79sqbAQ4MmvEAYqao",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "Qualified vLEI Issuer Credential",
            "description": "A vLEI Credential issued by GLEIF to Qualified vLEI Issuers which allows the Qualified "
            "vLEI Issuers to issue, verify and revoke Legal Entity vLEI Credentials and Legal "
            "Entity Official Organizational Role vLEI Credentials",
            "type": "object",
            "credentialType": "QualifiedvLEIIssuervLEICredential",
            "version": "1.0.0",
            "properties": {
                "v": {"description": "Version", "type": "string"},
                "d": {"description": "Credential SAID", "type": "string"},
                "u": {"description": "One time use nonce", "type": "string"},
                "i": {"description": "GLEIF Issuee AID", "type": "string"},
                "ri": {"description": "Credential status registry", "type": "string"},
                "s": {"description": "Schema SAID", "type": "string"},
                "a": {
                    "oneOf": [
                        {"description": "Attributes block SAID", "type": "string"},
                        {
                            "$id": "ELGgI0fkloqKWREXgqUfgS0bJybP1LChxCO3sqPSFHCj",
                            "description": "Attributes block",
                            "type": "object",
                            "properties": {
                                "d": {
                                    "description": "Attributes block SAID",
                                    "type": "string",
                                },
                                "i": {
                                    "description": "QVI Issuee AID",
                                    "type": "string",
                                },
                                "dt": {
                                    "description": "Issuance date time",
                                    "type": "string",
                                    "format": "date-time",
                                },
                                "LEI": {
                                    "description": "LEI of the requesting Legal Entity",
                                    "type": "string",
                                    "format": "ISO 17442",
                                },
                                "gracePeriod": {
                                    "description": "Allocated grace period",
                                    "type": "integer",
                                    "default": 90,
                                },
                            },
                            "additionalProperties": False,
                            "required": ["i", "dt", "LEI"],
                        },
                    ]
                },
                "r": {
                    "oneOf": [
                        {"description": "Rules block SAID", "type": "string"},
                        {
                            "$id": "ECllqarpkZrSIWCb97XlMpEZZH3q4kc--FQ9mbkFMb_5",
                            "description": "Rules block",
                            "type": "object",
                            "properties": {
                                "d": {
                                    "description": "Rules block SAID",
                                    "type": "string",
                                },
                                "usageDisclaimer": {
                                    "description": "Usage Disclaimer",
                                    "type": "object",
                                    "properties": {
                                        "l": {
                                            "description": "Associated legal language",
                                            "type": "string",
                                            "const": "Usage of a valid, unexpired, and non-revoked "
                                            "vLEI Credential, as defined in the "
                                            "associated Ecosystem Governance Framework, "
                                            "does not assert that the Legal Entity is "
                                            "trustworthy, honest, reputable in its "
                                            "business dealings, safe to do business with, "
                                            "or compliant with any laws or that an "
                                            "implied or expressly intended purpose will "
                                            "be fulfilled.",
                                        }
                                    },
                                },
                                "issuanceDisclaimer": {
                                    "description": "Issuance Disclaimer",
                                    "type": "object",
                                    "properties": {
                                        "l": {
                                            "description": "Associated legal language",
                                            "type": "string",
                                            "const": "All information in a valid, unexpired, "
                                            "and non-revoked vLEI Credential, as defined "
                                            "in the associated Ecosystem Governance "
                                            "Framework, is accurate as of the date the "
                                            "validation process was complete. The vLEI "
                                            "Credential has been issued to the legal "
                                            "entity or person named in the vLEI "
                                            "Credential as the subject; and the qualified "
                                            "vLEI Issuer exercised reasonable care to "
                                            "perform the validation process set forth in "
                                            "the vLEI Ecosystem Governance Framework.",
                                        }
                                    },
                                },
                            },
                            "additionalProperties": False,
                            "required": ["d", "usageDisclaimer", "issuanceDisclaimer"],
                        },
                    ]
                },
            },
            "additionalProperties": False,
            "required": ["i", "ri", "s", "d"],
        }
        _, sad = coring.Saider.saidify(sad, label=coring.Saids.dollar)
        schemer = scheming.Schemer(sed=sad)
        # NEW: "EBfdlu8R27Fbx-ehrqwImnK-8Cm79sqbAQ4MmvEAYqao"
        db.schema.pin(schemer.said, schemer)

        # OLD: "E1MCiPag0EWlqeJGzDA9xxr1bUSUR4fZXtqHDrwdXgbk"
        sad = {
            "$id": "",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "Legal Entity vLEI Credential",
            "description": "A vLEI Credential issued by a Qualified vLEI issuer to a Legal Entity",
            "credentialType": "LegalEntityvLEICredential",
            "properties": {
                "v": {"type": "string"},
                "d": {"type": "string"},
                "i": {"type": "string"},
                "ri": {"description": "credential status registry", "type": "string"},
                "s": {"description": "schema SAID", "type": "string"},
                "a": {
                    "description": "data block",
                    "properties": {
                        "d": {"type": "string"},
                        "i": {"type": "string"},
                        "dt": {
                            "description": "issuance date time",
                            "format": "date-time",
                            "type": "string",
                        },
                        "LEI": {"type": "string"},
                    },
                    "additionalProperties": False,
                    "required": ["i", "dt", "LEI"],
                    "type": "object",
                },
                "e": {"description": "edges block", "type": "object"},
                "r": {"type": "object", "description": "rules block"},
            },
            "additionalProperties": False,
            "required": ["i", "ri", "s", "d", "e", "r"],
            "type": "object",
        }

        _, sad = coring.Saider.saidify(sad, label=coring.Saids.dollar)
        schemer = scheming.Schemer(sed=sad)
        # NEW: "ENTAoj2oNBFpaniRswwPcca9W1ElEeH2V7ahw68HV4G5
        db.schema.pin(schemer.said, schemer)

        # OLD: "ExBYRwKdVGTWFq1M3IrewjKRhKusW9p9fdsdD0aSTWQI"
        sad = {
            "$id": "",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "GLEIF vLEI Credential",
            "description": "The vLEI Credential issued to GLEIF",
            "credentialType": "GLEIFvLEICredential",
            "type": "object",
            "properties": {
                "v": {"type": "string"},
                "d": {"type": "string"},
                "i": {"type": "string"},
                "ri": {"description": "credential status registry", "type": "string"},
                "s": {"description": "schema SAID", "type": "string"},
                "a": {
                    "description": "data block",
                    "properties": {
                        "d": {"type": "string"},
                        "i": {"type": "string"},
                        "dt": {
                            "description": "issuance date time",
                            "format": "date-time",
                            "type": "string",
                        },
                        "LEI": {"type": "string"},
                    },
                    "additionalProperties": False,
                    "required": ["d", "dt", "LEI"],
                    "type": "object",
                },
                "e": {"type": "object"},
            },
            "additionalProperties": False,
            "required": ["d", "i", "ri"],
        }
        _, sad = coring.Saider.saidify(sad, label=coring.Saids.dollar)
        schemer = scheming.Schemer(sed=sad)
        # NEW: EMQWEcCnVRk1hatTNyK3sIykYSrrFvafX3bHQ9Gkk1kC
        db.schema.pin(schemer.said, schemer)

        # OLD: EPz3ZvjQ_8ZwRKzfA5xzbMW8v8ZWLZhvOn2Kw1Nkqo_Q
        sad = {
            "$id": "",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "Legal Entity vLEI Credential",
            "description": "A vLEI Credential issued by a Qualified vLEI issuer to a Legal Entity",
            "credentialType": "LegalEntityvLEICredential",
            "properties": {
                "v": {"type": "string"},
                "d": {"type": "string"},
                "i": {"type": "string"},
                "ri": {"description": "credential status registry", "type": "string"},
                "s": {"description": "schema SAID", "type": "string"},
                "a": {
                    "description": "data block",
                    "properties": {
                        "d": {"type": "string"},
                        "i": {"type": "string"},
                        "dt": {
                            "description": "issuance date time",
                            "format": "date-time",
                            "type": "string",
                        },
                        "LEI": {"type": "string"},
                    },
                    "additionalProperties": False,
                    "required": ["i", "dt", "LEI"],
                    "type": "object",
                },
                "e": {
                    "description": "edges block",
                    "properties": {
                        "d": {"description": "SAID of edges block", "type": "string"},
                        "qualifiedvLEIIssuervLEICredential": {
                            "description": "node SAID of issuer credential",
                            "properties": {"n": {"type": "string"}},
                            "additionalProperties": False,
                            "required": ["n"],
                            "type": "object",
                        },
                    },
                    "additionalProperties": False,
                    "required": ["d", "qualifiedvLEIIssuervLEICredential"],
                    "type": "object",
                },
                "r": {
                    "type": "array",
                    "items": {"type": "object"},
                    "description": "rules block",
                    "minItems": 0,
                },
            },
            "additionalProperties": False,
            "required": ["i", "ri", "s", "d", "e", "r"],
            "type": "object",
        }
        _, sad = coring.Saider.saidify(sad, label=coring.Saids.dollar)
        schemer = scheming.Schemer(sed=sad)
        # NEW: ED892b40P_GcESs3wOcc2zFvL_GVi2Ybzp9isNTZKqP0
        db.schema.pin(schemer.said, schemer)

        # OLD: EBfdlu8R27Fbx-ehrqwImnK-8Cm79sqbAQ4MmvEAYqao
        sad = {
            "$id": "",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "Qualified vLEI Issuer Credential",
            "description": "A vLEI Credential issued by GLEIF to Qualified vLEI Issuers which allows the Qualified "
            "vLEI Issuers to issue, verify and revoke Legal Entity vLEI Credentials and Legal "
            "Entity Official Organizational Role vLEI Credentials",
            "credentialType": "QualifiedvLEIIssuervLEICredential",
            "properties": {
                "v": {"type": "string"},
                "d": {"type": "string"},
                "i": {"type": "string"},
                "ri": {"description": "credential status registry", "type": "string"},
                "s": {"description": "schema SAID", "type": "string"},
                "a": {
                    "description": "data block",
                    "properties": {
                        "d": {"type": "string"},
                        "i": {"type": "string"},
                        "dt": {
                            "description": "issuance date time",
                            "format": "date-time",
                            "type": "string",
                        },
                        "LEI": {"type": "string"},
                        "gracePeriod": {"default": 90, "type": "integer"},
                    },
                    "additionalProperties": False,
                    "required": ["i", "dt", "LEI"],
                    "type": "object",
                },
                "e": {"type": "object"},
            },
            "additionalProperties": False,
            "required": ["i", "ri", "s", "d"],
            "type": "object",
        }

        _, sad = coring.Saider.saidify(sad, label=coring.Saids.dollar)
        schemer = scheming.Schemer(sed=sad)
        # NEW: EFgnk_c08WmZGgv9_mpldibRuqFMTQN-rAgtD-TCOwbs
        db.schema.pin(schemer.said, schemer)


class Helpers:
    controllerAID = "EK35JRNdfVkO4JwhXaSTdV4qzB_ibk_tGJmSVcY4pZqx"

    @staticmethod
    def remove_test_dirs(name, kdir="/usr/local/var/keri"):
        # if os.path.exists(f'{kdir}/adb/TheAgency'):
        #     shutil.rmtree(f'{kdir}/adb/TheAgency')
        if os.path.exists(f"{kdir}/cf/{name}.json"):
            os.remove(f"{kdir}/cf/{name}.json")
        if os.path.exists(f"{kdir}/cf/{name}"):
            shutil.rmtree(f"{kdir}/cf/{name}")
        if os.path.exists(f"{kdir}/db/{name}"):
            shutil.rmtree(f"{kdir}/db/{name}")
        if os.path.exists(f"{kdir}/ks/{name}"):
            shutil.rmtree(f"{kdir}/ks/{name}")
        if os.path.exists(f"{kdir}/mbx/{name}"):
            shutil.rmtree(f"{kdir}/mbx/{name}")
        if os.path.exists(f"{kdir}/not/{name}"):
            shutil.rmtree(f"{kdir}/not/{name}")
        if os.path.exists(f"{kdir}/opr/{name}"):
            shutil.rmtree(f"{kdir}/opr/{name}")
        if os.path.exists(f"{kdir}/reg/{name}"):
            shutil.rmtree(f"{kdir}/reg/{name}")
        if os.path.exists(f"{kdir}/reg/agent-{name}"):
            shutil.rmtree(f"{kdir}/reg/agent-{name}")
        if os.path.exists(f"{kdir}/rks/{name}"):
            shutil.rmtree(f"{kdir}/rks/{name}")

    @staticmethod
    def server(agency, httpPort=3902):
        app = falcon.App()
        httpEnd = indirecting.HttpEnd(agency=agency)
        app.add_route("/", httpEnd)
        server = http.Server(port=httpPort, app=app)
        httpServerDoer = http.ServerDoer(server=server)
        return httpServerDoer

    @staticmethod
    def controller(bran=b"0123456789abcdefghijk"):
        """
        Creates a Signify Controller using a Signify stem in the key seed using a single signer.
        Returns:
            (SerderKERI, List[Siger]): The inception event serder and list of signatures
        """
        serder, signers = Helpers.incept(bran=bran, stem="signify:controller", pidx=0)
        sigers = [signers[0].sign(ser=serder.raw, index=0)]
        return serder, sigers

    @staticmethod
    def incept(
        bran, stem, pidx, count=1, sith="1", wits=None, toad="0", delpre=None
    ) -> Tuple[serdering.SerderKERI, List[Signer]]:
        """
        Perform a regular inception event with the given parameters.
        Returns:
            (SerderKERI, List[Signer]): The inception event serder and list of signers
        """
        if wits is None:
            wits = []

        salter = core.Salter(raw=bran)
        creator = keeping.SaltyCreator(
            salt=salter.qb64, stem=stem, tier=coring.Tiers.low
        )

        signers = creator.create(
            pidx=pidx, ridx=0, tier=coring.Tiers.low, temp=False, count=count
        )
        nsigners = creator.create(
            pidx=pidx, ridx=1, tier=coring.Tiers.low, temp=False, count=count
        )

        keys = [signer.verfer.qb64 for signer in signers]
        ndigs = [coring.Diger(ser=nsigner.verfer.qb64b) for nsigner in nsigners]

        if delpre is not None:
            serder = eventing.delcept(
                delpre=delpre,
                keys=keys,
                isith=sith,
                nsith=sith,
                ndigs=[diger.qb64 for diger in ndigs],
                code=coring.MtrDex.Blake3_256,
                toad=toad,
                wits=wits,
            )
        else:
            serder = eventing.incept(
                keys=keys,
                isith=sith,
                nsith=sith,
                ndigs=[diger.qb64 for diger in ndigs],
                code=coring.MtrDex.Blake3_256,
                toad=toad,
                wits=wits,
            )

        return serder, signers

    @staticmethod
    def inceptRandy(bran, count=1, sith="1", wits=None, toad="0"):
        if wits is None:
            wits = []

        salter = core.Salter(raw=bran)
        signer = salter.signer(transferable=False)
        aeid = signer.verfer.qb64
        encrypter = core.Encrypter(verkey=aeid)

        creator = keeping.RandyCreator()
        signers = creator.create(count=count)
        prxs = [encrypter.encrypt(prim=signer).qb64 for signer in signers]
        nsigners = creator.create(count=count)
        nxts = [encrypter.encrypt(prim=signer).qb64 for signer in nsigners]

        keys = [signer.verfer.qb64 for signer in signers]
        ndigs = [coring.Diger(ser=nsigner.verfer.qb64b) for nsigner in nsigners]

        serder = eventing.incept(
            keys=keys,
            isith=sith,
            nsith=sith,
            ndigs=[diger.qb64 for diger in ndigs],
            code=coring.MtrDex.Blake3_256,
            toad=toad,
            wits=wits,
        )

        return serder, signers, prxs, nxts

    @staticmethod
    def inceptExtern(count=1, sith="1", wits=None, toad="0"):
        if wits is None:
            wits = []

        creator = keeping.RandyCreator()
        signers = creator.create(count=count)
        nsigners = creator.create(count=count)

        keys = [signer.verfer.qb64 for signer in signers]
        ndigs = [coring.Diger(ser=nsigner.verfer.qb64b) for nsigner in nsigners]

        serder = eventing.incept(
            keys=keys,
            isith=sith,
            nsith=sith,
            ndigs=[diger.qb64 for diger in ndigs],
            code=coring.MtrDex.Blake3_256,
            toad=toad,
            wits=wits,
        )

        return serder, signers

    @staticmethod
    def interact(pre, bran, pidx, ridx, sn, dig, data):
        serder = eventing.interact(pre=pre, dig=dig, sn=sn, data=data)
        salter = core.Salter(raw=bran)
        creator = keeping.SaltyCreator(
            salt=salter.qb64, stem="signify:aid", tier=coring.Tiers.low
        )

        signers = creator.create(
            pidx=pidx, ridx=ridx, tier=coring.Tiers.low, temp=False, count=1
        )
        sigers = [signer.sign(ser=serder.raw, index=0).qb64 for signer in signers]
        return serder, sigers

    @staticmethod
    def createRotate(aid, salt, signers, pidx, ridx, kidx, wits, toad):
        salter = core.Salter(raw=salt)
        creator = keeping.SaltyCreator(
            salt=salter.qb64, stem="signify:aid", tier=coring.Tiers.low
        )
        encrypter = core.Encrypter(verkey=signers[0].verfer.qb64)
        sxlt = encrypter.encrypt(ser=salter.qb64).qb64

        rsigners = creator.create(
            pidx=pidx, ridx=ridx, tier=coring.Tiers.low, temp=False, count=1
        )
        rnsigners = creator.create(
            pidx=pidx, ridx=ridx + 1, tier=coring.Tiers.low, temp=False, count=1
        )

        rkeys = [signer.verfer.qb64 for signer in rsigners]
        rndigs = [coring.Diger(ser=nsigner.verfer.qb64b) for nsigner in rnsigners]

        serder = eventing.rotate(
            pre=aid["prefix"],
            keys=rkeys,
            dig=aid["prefix"],
            ndigs=[diger.qb64 for diger in rndigs],
            wits=wits,
            toad=toad,
        )
        sigers = [signer.sign(ser=serder.raw, index=0).qb64 for signer in rsigners]
        body = {
            "rot": serder.ked,
            "sigs": sigers,
            "salty": {
                "stem": "signify:aid",
                "pidx": pidx,
                "tier": "low",
                "sxlt": sxlt,
                "transferable": True,
                "kidx": kidx,
                "icodes": [MtrDex.Ed25519_Seed],
                "ncodes": [MtrDex.Ed25519_Seed],
            },
        }
        return body

    @staticmethod
    def sign(bran, pidx, ridx, ser):
        salter = core.Salter(raw=bran)
        creator = keeping.SaltyCreator(
            salt=salter.qb64, stem="signify:aid", tier=coring.Tiers.low
        )

        signers = creator.create(
            pidx=pidx, ridx=ridx, tier=coring.Tiers.low, temp=False, count=1
        )
        sigers = [signer.sign(ser=ser, index=0).qb64 for signer in signers]
        return sigers

    @staticmethod
    def createAid(client, name, salt, wits=None, toad="0", delpre=None):
        serder, signers = Helpers.incept(
            salt, "signify:aid", pidx=0, wits=wits, toad=toad, delpre=delpre
        )
        assert len(signers) == 1

        salter = core.Salter(raw=salt)
        encrypter = core.Encrypter(verkey=signers[0].verfer.qb64)
        sxlt = encrypter.encrypt(ser=salter.qb64).qb64

        sigers = [signer.sign(ser=serder.raw, index=0).qb64 for signer in signers]

        body = {
            "name": name,
            "icp": serder.ked,
            "sigs": sigers,
            "salty": {
                "stem": "signify:aid",
                "pidx": 0,
                "tier": "low",
                "sxlt": sxlt,
                "transferable": True,
                "kidx": 0,
                "icodes": [MtrDex.Ed25519_Seed],
                "ncodes": [MtrDex.Ed25519_Seed],
            },
        }

        res = client.simulate_post(path="/identifiers", body=json.dumps(body))
        assert res.status_code == 200 or res.status_code == 202
        return res.json

    @staticmethod
    def createEndRole(client, agent, recp, name, salt):
        rpy = Helpers.endrole(recp, agent.agentHab.pre)
        sigs = Helpers.sign(salt, 0, 0, rpy.raw)
        body = dict(rpy=rpy.ked, sigs=sigs)

        res = client.simulate_post(path=f"/identifiers/{name}/endroles", json=body)
        op = res.json
        ked = op["response"]
        serder = serdering.SerderKERI(sad=ked)
        assert serder.raw == rpy.raw

    @staticmethod
    def createRegistry(client, agent, salt, doist, deeds):
        op = Helpers.createAid(client, "issuer", salt)
        aid = op["response"]
        pre = aid["i"]
        assert pre == "EHgwVwQT15OJvilVvW57HE4w0-GPs_Stj2OFoAHZSysY"

        nonce = Salter().qb64
        regser = veventing.incept(
            pre,
            baks=[],
            toad="0",
            nonce=nonce,
            cnfg=[TraitCodex.NoBackers],
            code=coring.MtrDex.Blake3_256,
        )
        anchor = dict(i=regser.ked["i"], s=regser.ked["s"], d=regser.said)
        serder, sigers = Helpers.interact(
            pre=pre, bran=salt, pidx=0, ridx=0, dig=aid["d"], sn="1", data=[anchor]
        )
        body = dict(
            name="test", alias="test", vcp=regser.ked, ixn=serder.ked, sigs=sigers
        )
        result = client.simulate_post(
            path="/identifiers/issuer/registries", body=json.dumps(body).encode("utf-8")
        )
        op = result.json
        metadata = op["metadata"]

        assert op["done"] is True
        assert metadata["anchor"] == anchor
        assert result.status == falcon.HTTP_202

        while regser.pre not in agent.tvy.tevers:
            doist.recur(deeds=deeds)

        assert regser.pre in agent.tvy.tevers

        result = client.simulate_get(path="/identifiers/issuer/registries")
        registries = result.json
        assert len(registries) == 1
        assert registries[0]["name"] == "test"

        issuer = client.simulate_get("/identifiers/issuer")

        return registries[0], issuer.json

    @staticmethod
    def endrole(cid, eid, role="agent"):
        data = dict(cid=cid, role=role, eid=eid)
        return eventing.reply(route="/end/role/add", data=data)

    @staticmethod
    def locscheme(eid, url, scheme="http"):
        data = dict(eid=eid, url=url, scheme=scheme)
        return eventing.reply(route="/loc/scheme", data=data)

    @staticmethod
    def middleware(agent):
        return MockAgentMiddleware(agent=agent)

    @staticmethod
    @contextmanager
    def openKeria(salter=None, cf=None, temp=True):
        """
        Creates a Signify Controller and it's associated Agent, Agent Habery, containing Agency,
        Falcon HTTP App, and Falcon test client. The Agency has a single Agent.

        Returns:
            (Agency, Agent, falcon.App, TestClient): The agency, agent, Falcon HTTP app, and Falcon HTTP test client
        """
        serder, sigers = Helpers.controller()
        assert serder.pre == Helpers.controllerAID

        if salter is None:
            salt = b"0123456789abcdef"
            salter = core.Salter(raw=salt)

        if cf is None:
            cf = configing.Configer(
                name="keria", headDirPath=SCRIPTS_DIR, reopen=True, clear=False
            )

        agency = agenting.Agency(name="agency", bran=None, temp=True)
        with habbing.openHby(name="keria", salt=salter.qb64, temp=temp, cf=cf) as hby:
            ims = eventing.messagize(serder, sigers=sigers)
            parsing.Parser(kvy=hby.kvy).parseOne(ims=ims)

            agentHab = hby.makeHab(
                serder.pre, ns="agent", transferable=True, delpre=serder.pre
            )

            rgy = credentialing.Regery(
                hby=hby, name=agentHab.name, base=hby.base, temp=True
            )
            agent = agenting.Agent(
                hby=hby, rgy=rgy, agentHab=agentHab, agency=agency, caid=serder.pre
            )
            agency.agents[serder.pre] = agent

            app = falcon.App()
            app.add_middleware(Helpers.middleware(agent))
            client = testing.TestClient(app)
            yield agency, agent, app, client

    @staticmethod
    @contextmanager
    def withIssuer(name, hby):
        yield Issuer(name=name, hby=hby)

    @staticmethod
    def mockNowUTC():
        """
        Use predetermined value for now (current time)
        '2021-01-01T00:00:00.000000+00:00'
        """
        return helping.fromIso8601("2021-01-01T00:00:00.000000+00:00")

    @staticmethod
    def mockNowIso8601():
        """
        Use predetermined value for now (current time)
        '2021-01-01T00:00:00.000000+00:00'
        """
        return "2021-06-27T21:26:21.233257+00:00"

    @staticmethod
    def mockRandomNonce():
        return "A9XfpxIl1LcIkMhUSCCC8fgvkuX8gG9xK3SM-S8a8Y_U"

    @staticmethod
    def witnessMsg(hab, msg, sn, witHabs):
        rctMsgs = []
        for i, witHab in enumerate(witHabs):
            kvy = witHab.kvy
            witHab.psr.parse(ims=bytearray(msg), kvy=kvy, local=True)
            # accepted event with cam sigs since own witness
            assert kvy.kevers[hab.pre].sn == sn
            assert len(kvy.cues) >= 1  # at least queued receipt cue
            # better to find receipt cue in cues exactly
            rctMsg = witHab.processCues(kvy.cues)  # process cue returns rct msg
            assert len(rctMsg) > len(msg)
            rctMsgs.append(rctMsg)

        for rMsg in rctMsgs:  # process rct msgs from all witnesses
            hab.psr.parse(ims=bytearray(rMsg), kvy=hab.kvy, local=True)
        for whab in witHabs:
            assert whab.pre in hab.kvy.kevers

        return rctMsgs


class Issuer:
    LE = "ENTAoj2oNBFpaniRswwPcca9W1ElEeH2V7ahw68HV4G5"
    QVI = "EFgnk_c08WmZGgv9_mpldibRuqFMTQN-rAgtD-TCOwbs"
    date = "2021-06-27T21:26:21.233257+00:00"

    def __init__(self, name, hby):
        self.hby = hby
        self.rgy = Regery(hby=hby, name=name, temp=True)
        self.registrar = Registrar(hby=hby, rgy=self.rgy, counselor=None)
        self.verifier = verifying.Verifier(hby=hby, reger=self.rgy.reger)

    def createRegistry(self, pre, name):
        conf = dict(nonce="AGu8jwfkyvVXQ2nqEb5yVigEtR31KSytcpe2U2f7NArr")
        hab = self.hby.habs[pre]

        registry = self.rgy.makeRegistry(name=name, prefix=pre, **conf)
        assert registry.regk == "EACehJRd0wfteUAJgaTTJjMSaQqWvzeeHqAMMqxuqxU4"

        rseal = SealEvent(registry.regk, "0", registry.regd)
        rseal = dict(i=rseal.i, s=rseal.s, d=rseal.d)
        anc = hab.interact(data=[rseal])

        aserder = serdering.SerderKERI(raw=bytes(anc))
        self.registrar.incept(iserder=registry.vcp, anc=aserder)

        # Process escrows to clear event
        self.rgy.processEscrows()
        self.registrar.processEscrows()

        assert (
            self.rgy.reger.ctel.get(keys=(registry.regk, coring.Seqner(sn=0).qb64))
            is not None
        )

    def issueLegalEntityvLEI(self, reg, issuer, issuee, LEI):
        registry = self.rgy.registryByName(reg)
        credSubject = dict(
            d="",
            i=issuee,
            dt=self.date,
            LEI=LEI,
        )

        # Create the Creder and issue the credentials
        creder = proving.credential(
            issuer=issuer.pre,
            schema=self.LE,
            data=credSubject,
            status=registry.regk,
            source={},
            rules={},
        )

        iserder = registry.issue(said=creder.said, dt=self.date)

        vcid = iserder.ked["i"]
        rseq = coring.Seqner(snh=iserder.ked["s"])
        rseal = eventing.SealEvent(vcid, rseq.snh, iserder.said)
        rseal = dict(i=rseal.i, s=rseal.s, d=rseal.d)

        anc = issuer.interact(data=[rseal])
        aserder = serdering.SerderKERI(raw=anc)
        self.registrar.issue(creder=creder, iserder=iserder, anc=aserder)

        prefixer = coring.Prefixer(qb64=iserder.pre)
        seqner = coring.Seqner(sn=iserder.sn)
        saider = coring.Saider(qb64=iserder.said)
        craw = signing.serialize(creder, prefixer, seqner, saider)

        # Process escrows to clear event
        self.rgy.processEscrows()
        self.registrar.processEscrows()

        parsing.Parser().parse(ims=craw, vry=self.verifier)

        assert self.rgy.reger.saved.get(keys=creder.said) is not None

        return creder.said

    def issueQVIvLEI(self, reg, issuer, issuee, LEI):
        registry = self.rgy.registryByName(reg)
        credSubject = dict(
            d="",
            i=issuee,
            dt=self.date,
            LEI=LEI,
        )

        # Create the Creder and issue the credentials
        creder = proving.credential(
            issuer=issuer.pre, schema=self.QVI, data=credSubject, status=registry.regk
        )

        iserder = registry.issue(said=creder.said, dt=self.date)

        vcid = iserder.ked["i"]
        rseq = coring.Seqner(snh=iserder.ked["s"])
        rseal = eventing.SealEvent(vcid, rseq.snh, iserder.said)
        rseal = dict(i=rseal.i, s=rseal.s, d=rseal.d)

        anc = issuer.interact(data=[rseal])
        aserder = serdering.SerderKERI(raw=anc)
        self.registrar.issue(creder=creder, iserder=iserder, anc=aserder)

        prefixer = coring.Prefixer(qb64=iserder.pre)
        seqner = coring.Seqner(sn=iserder.sn)
        saider = coring.Saider(qb64=iserder.said)
        craw = signing.serialize(creder, prefixer, seqner, saider)

        # Process escrows to clear event
        self.rgy.processEscrows()
        self.registrar.processEscrows()

        parsing.Parser().parse(ims=craw, vry=self.verifier)

        assert self.rgy.reger.saved.get(keys=creder.said) is not None

        return creder.said


class MockAgentMiddleware:
    def __init__(self, agent):
        self.agent = agent

    def process_request(self, req, _):
        """Process request to ensure has a valid signature from caid

        Parameters:
            req: Http request object
            _: placeholder for Http response object


        """
        req.context.agent = self.agent
